# Filename: calib.py
# pylint: disable=locally-disabled
"""
Calibration.

"""
import numpy as np

from .db import DBManager
from .core import Module
from .hardware import Detector
from .dataclasses import Table
from .tools import istype
from .logger import get_logger

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = ["Thomas Heid"]
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

log = get_logger(__name__)

try:
    import numba as nb
except (ImportError, OSError):
    HAVE_NUMBA = False
    jit = lambda f: f
    log.warning(
        "No numba detected, consider `pip install numba` for more speed!"
    )
else:
    try:
        from numba.typed import Dict
    except ImportError:
        log.warning("Please update numba (0.43+) to have dictionary support!")
        HAVE_NUMBA = False
        jit = lambda f: f
    else:
        HAVE_NUMBA = True
        from numba import jit


class Calibration(Module):
    """A very simple, preliminary Module which gives access to the calibration.

    Parameters
    ----------
    apply: bool, optional [default=True]
        Apply the calibration to the hits (add position/direction/t0)?
    filename: str, optional [default=None]
        DetX file with detector description.
    det_id: int, optional
        .detx ID of detector (when retrieving from database).
    t0set: optional
        t0set (when retrieving from database).
    calibset: optional
        calibset (when retrieving from database).
    """
    __name__ = 'Calibration'
    name = 'Calibration'

    def configure(self):
        self._should_apply = self.get('apply', default=True)
        self.filename = self.get('filename')
        self.det_id = self.get('det_id')
        self.run = self.get('run')
        self.t0set = self.get('t0set')
        self.calibset = self.get('calibset')
        self.detector = self.get('detector')
        self._pos_dom_channel = None
        self._dir_dom_channel = None
        self._t0_dom_channel = None
        self._pos_pmt_id = None
        self._dir_pmt_id = None
        self._t0_pmt_id = None
        self._lookup_tables = None    # for Numba

        if self.det_id and self.run:
            self.cprint(
                "Grabbing the calibration for Det ID {} and run {}".format(
                    self.det_id, self.run
                )
            )
            raw_detx = DBManager().detx_for_run(self.det_id, self.run)
            self.detector = Detector(string=raw_detx)
            self._create_dom_channel_lookup()
            self._create_pmt_id_lookup()
            return

        # TODO: deprecation
        if self.get('calibration'):
            self.log.warning(
                "The parameter 'calibration' has been renamed "
                "to 'calibset'. The 'calibration' parameter will be removed "
                "in the next version of KM3Pipe"
            )
            self.calibset = self.get('calibration')

        if self.filename or self.det_id:
            if self.filename is not None:
                self.detector = Detector(filename=self.filename)
            if self.det_id:
                self.detector = Detector(
                    det_id=self.det_id,
                    t0set=self.t0set,
                    calibration=self.calibset
                )

        if self.detector is not None:
            self.log.debug("Creating lookup tables")
            self._create_dom_channel_lookup()
            self._create_pmt_id_lookup()
        else:
            self.log.critical("No detector information loaded.")

    def process(self, blob, key='Hits', outkey='CalibHits'):
        if self._should_apply:
            blob[outkey] = self.apply(blob[key])
        return blob

    def get_detector(self):
        """Return the detector"""
        return self.detector

    def apply_t0(self, hits):
        """Apply only t0s"""
        if HAVE_NUMBA:
            apply_t0_nb(
                hits.time, hits.dom_id, hits.channel_id, self._lookup_tables
            )
        else:
            n = len(hits)
            cal = np.empty(n)
            lookup = self._calib_by_dom_and_channel
            for i in range(n):
                calib = lookup[hits['dom_id'][i]][hits['channel_id'][i]]
                cal[i] = calib[6]
            hits.time += cal
        return hits

    def apply(self, hits, no_copy=False, correct_slewing=False):
        """Add x, y, z, t0 (and du, floor if DataFrame) columns to the hits.

        """
        if not no_copy:
            hits = hits.copy()

        if istype(hits, 'DataFrame'):
            # do we ever see McHits here?
            hits = Table.from_template(hits, 'Hits')

        if hasattr(hits, 'dom_id') and hasattr(hits, 'channel_id'):
            dir_x, dir_y, dir_z, du, floor, pos_x, pos_y, pos_z, t0 = _get_calibration_for_hits(
                hits, self._calib_by_dom_and_channel
            )
        elif hasattr(hits, 'pmt_id'):
            dir_x, dir_y, dir_z, du, floor, pos_x, pos_y, pos_z, t0 = _get_calibration_for_mchits(
                hits, self._calib_by_pmt_id
            )
        else:
            raise TypeError(
                "Don't know how to apply calibration to '{0}'. "
                "We need at least 'dom_id' and 'channel_id', or "
                "'pmt_id'.".format(hits.name)
            )

        if hasattr(hits, 'time'):
            if hits.time.dtype != t0.dtype:
                time = hits.time.astype('f4') + t0.astype('f4')
                hits = hits.drop_columns(['time'])
                hits = hits.append_columns(['time'], [time])
            else:
                hits.time += t0

        hits_data = {}
        for colname in hits.dtype.names:
            hits_data[colname] = hits[colname]
        calib = {
            'dir_x': dir_x,
            'dir_y': dir_y,
            'dir_z': dir_z,
            'du': du.astype(np.uint8),
            'floor': du.astype(np.uint8),
            'pos_x': pos_x,
            'pos_y': pos_y,
            'pos_z': pos_z,
            't0': t0,
        }
        hits_data.update(calib)
        if correct_slewing:
            hits_data['time'] -= slew(hits_data['tot'])
        return Table(
            hits_data,
            h5loc=hits.h5loc,
            split_h5=hits.split_h5,
            name=hits.name
        )

    def _create_dom_channel_lookup(self):
        if HAVE_NUMBA:
            from numba.typed import Dict
            from numba import types
            data = Dict.empty(
                key_type=types.i8, value_type=types.float64[:, :]
            )
        else:
            data = {}
        for pmt in self.detector.pmts:
            if pmt.dom_id not in data:
                data[pmt.dom_id] = np.zeros((31, 9))
            data[pmt.dom_id][pmt.channel_id] = np.asarray([
                pmt.pos_x, pmt.pos_y, pmt.pos_z, pmt.dir_x, pmt.dir_y,
                pmt.dir_z, pmt.t0, pmt.du, pmt.floor
            ],
                                                          dtype=np.float64)
        self._calib_by_dom_and_channel = data
        if HAVE_NUMBA:
            self._lookup_tables = [(dom, cal) for dom, cal in data.items()]

    def _create_pmt_id_lookup(self):
        if HAVE_NUMBA:
            from numba.typed import Dict
            from numba import types
            data = Dict.empty(key_type=types.i8, value_type=types.float64[:])
        else:
            data = {}
        for pmt in self.detector.pmts:
            data[pmt.pmt_id] = np.asarray([
                pmt.pos_x,
                pmt.pos_y,
                pmt.pos_z,
                pmt.dir_x,
                pmt.dir_y,
                pmt.dir_z,
                pmt.t0,
                pmt.du,
                pmt.floor,
            ],
                                          dtype=np.float64)
        self._calib_by_pmt_id = data

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return "Calibration: det_id({0})".format(self.det_id)


@jit
def apply_t0_nb(times, dom_ids, channel_ids, lookup_tables):
    """Apply t0s using a lookup table of tuples (dom_id, calib)"""
    dom_id = 0
    lookup = np.empty((31, 9))
    for i in range(len(times)):
        cur_dom_id = dom_ids[i]
        if cur_dom_id != dom_id:
            dom_id = cur_dom_id
            for (d, m) in lookup_tables:
                if d == dom_id:
                    np.copyto(lookup, m)
        t0 = lookup[channel_ids[i]][6]
        times[i] += t0


@jit
def _get_calibration_for_hits(hits, lookup):
    """Append the position, direction and t0 columns and add t0 to time"""
    n = len(hits)
    cal = np.empty((n, 9))
    for i in range(n):
        calib = lookup[hits['dom_id'][i]][hits['channel_id'][i]]
        cal[i] = calib
    dir_x = cal[:, 3]
    dir_y = cal[:, 4]
    dir_z = cal[:, 5]
    du = cal[:, 7]
    floor = cal[:, 8]
    pos_x = cal[:, 0]
    pos_y = cal[:, 1]
    pos_z = cal[:, 2]

    t0 = cal[:, 6]

    return [dir_x, dir_y, dir_z, du, floor, pos_x, pos_y, pos_z, t0]


@jit
def _get_calibration_for_mchits(hits, lookup):
    """Append the position, direction and t0 columns and add t0 to time"""
    n_hits = len(hits)
    cal = np.empty((n_hits, 9))
    for i in range(n_hits):
        cal[i] = lookup[hits['pmt_id'][i]]
    dir_x = cal[:, 3]
    dir_y = cal[:, 4]
    dir_z = cal[:, 5]
    du = cal[:, 7]
    floor = cal[:, 8]
    pos_x = cal[:, 0]
    pos_y = cal[:, 1]
    pos_z = cal[:, 2]
    t0 = cal[:, 6]

    return [dir_x, dir_y, dir_z, du, floor, pos_x, pos_y, pos_z, t0]


class CalibrationService(Module):
    """A service which provides calibration routines for hits

    Parameters
    ----------
    filename: str, optional [default=None]
        DetX file with detector description.
    det_id: int, optional
        .detx ID of detector (when retrieving from database).
    t0set: optional
        t0set (when retrieving from database).
    calibset: optional
        calibset (when retrieving from database).
    detector: kp.hardware.Detector, optional
    """
    __name__ = 'Calibration'
    name = 'Calibration'

    def configure(self):
        self.filename = self.get('filename')
        self.det_id = self.get('det_id')
        self.t0set = self.get('t0set')
        self.calibset = self.get('calibset')

        self._detector = self.get('detector')

        if self._detector is not None:
            self._calibration = Calibration(detector=self._detector)

        self._calibration = None

        self.expose(self.calibrate, "calibrate")
        self.expose(self.get_detector, "get_detector")
        self.expose(self.get_calibration, "get_calibration")
        self.expose(self.load_calibration, "load_calibration")
        self.expose(self.correct_slewing, "correct_slewing")

        self.expose(self.detector_deprecation, "detector")

    def load_calibration(
        self, filename=None, det_id=None, t0set=None, calibset=None
    ):
        """Load another calibration"""
        self.filename = filename
        self.det_id = det_id
        self.t0set = t0set
        self.calibset = calibset
        self._detector = None
        self._calibration = None

    @property
    def detector_deprecation(self):
        self.log.deprecation(
            "The service 'detector' of the CalibrationService has been "
            "deprecated. Please use 'get_detector()' instead in future."
        )

    def calibrate(self, hits):
        return self.calibration.apply(hits)

    @property
    def detector(self):
        if self._detector is None:
            self._detector = self.calibration.detector
        return self._detector

    def get_detector(self):
        """Extra getter to be as lazy as possible (expose triggers otherwise"""
        return self.detector

    @property
    def calibration(self):
        if self._calibration is None:
            self._calibration = Calibration(
                filename=self.filename,
                det_id=self.det_id,
                t0set=self.t0set,
                calibset=self.calibset
            )
        return self._calibration

    def get_calibration(self):
        """Extra getter to be as lazy as possible (expose triggers otherwise"""
        return self.calibration

    def correct_slewing(self, hits):
        """Apply time slewing correction to the hit times"""
        hits.time -= slew(hits.tot)


@jit
def slew(tot):
    """Calculate the time slewing of a PMT response for a given ToT


    Parameters
    ----------
    tot: int or np.array(int)
      Time over threshold value of a hit

    Returns
    -------
    time: int or np.array(int)
      Time slewing, which has to be subtracted from the original hit time.
    """

    # First parametrisation
    #p0 = 7.70824
    #p1 = 0.00879447
    #p2 = -0.0621101
    #p3 = -1.90226

    # Second parametrisation
    # p0 =  13.6488662517;
    # p1 =  -0.128744123166;
    # p2 =  -0.0174837749244;
    # p3 =  -4.47119633965;

    # corr = p0 * np.exp(p1 * np.sqrt(tot) + p2 * tot) + p3
    #      this->push_back(  8.01,     #
    # return corr
    corr = np.array([
        8.01, 7.52, 7.05, 6.59, 6.15, 5.74, 5.33, 4.95, 4.58, 4.22, 3.89, 3.56,
        3.25, 2.95, 2.66, 2.39, 2.12, 1.87, 1.63, 1.40, 1.19, 0.98, 0.78, 0.60,
        0.41, 0.24, 0.07, -0.10, -0.27, -0.43, -0.59, -0.75, -0.91, -1.08,
        -1.24, -1.41, -1.56, -1.71, -1.85, -1.98, -2.11, -2.23, -2.35, -2.47,
        -2.58, -2.69, -2.79, -2.89, -2.99, -3.09, -3.19, -3.28, -3.37, -3.46,
        -3.55, -3.64, -3.72, -3.80, -3.88, -3.96, -4.04, -4.12, -4.20, -4.27,
        -4.35, -4.42, -4.49, -4.56, -4.63, -4.70, -4.77, -4.84, -4.90, -4.97,
        -5.03, -5.10, -5.16, -5.22, -5.28, -5.34, -5.40, -5.46, -5.52, -5.58,
        -5.63, -5.69, -5.74, -5.80, -5.85, -5.91, -5.96, -6.01, -6.06, -6.11,
        -6.16, -6.21, -6.26, -6.31, -6.36, -6.41, -6.45, -6.50, -6.55, -6.59,
        -6.64, -6.68, -6.72, -6.77, -6.81, -6.85, -6.89, -6.93, -6.98, -7.02,
        -7.06, -7.09, -7.13, -7.17, -7.21, -7.25, -7.28, -7.32, -7.36, -7.39,
        -7.43, -7.46, -7.50, -7.53, -7.57, -7.60, -7.63, -7.66, -7.70, -7.73,
        -7.76, -7.79, -7.82, -7.85, -7.88, -7.91, -7.94, -7.97, -7.99, -8.02,
        -8.05, -8.07, -8.10, -8.13, -8.15, -8.18, -8.20, -8.23, -8.25, -8.28,
        -8.30, -8.32, -8.34, -8.37, -8.39, -8.41, -8.43, -8.45, -8.47, -8.49,
        -8.51, -8.53, -8.55, -8.57, -8.59, -8.61, -8.62, -8.64, -8.66, -8.67,
        -8.69, -8.70, -8.72, -8.74, -8.75, -8.76, -8.78, -8.79, -8.81, -8.82,
        -8.83, -8.84, -8.86, -8.87, -8.88, -8.89, -8.90, -8.92, -8.93, -8.94,
        -8.95, -8.96, -8.97, -8.98, -9.00, -9.01, -9.02, -9.04, -9.04, -9.04,
        -9.04, -9.04, -9.04, -9.04, -9.04, -9.04, -9.04, -9.04, -9.04, -9.04,
        -9.04, -9.04, -9.04, -9.04, -9.04, -9.04, -9.04, -9.04, -9.04, -9.04,
        -9.04, -9.04, -9.04, -9.04, -9.04, -9.04, -9.04, -9.04, -9.04, -9.04,
        -9.04, -9.04, -9.04, -9.04, -9.04, -9.04, -9.04, -9.04, -9.04, -9.04,
        -9.04, -9.04, -9.04, -9.04, -9.04, -9.04, -9.04, -9.04, -9.04, -9.04,
        -9.04, -9.04
    ])
    return corr[tot]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Tamas Gal <tgal@km3net.de>
# License: MIT
#!/usr/bin/env python
# vim: ts=4 sw=4 et
"""
=============================
UDP Forwarder for ControlHost
=============================

A simple UDP forwarder for ControlHost messages.

This application is used to forward monitoring channel data from Ligier
to a given UDP address.

"""

import socket
import sys

import km3pipe as kp

__author__ = "Tamas Gal"
__email__ = "tgal@km3net.de"


class UDPForwarder(kp.Module):
    def configure(self):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.counter = 0

    def process(self, blob):
        if str(blob['CHPrefix'].tag) == 'IO_MONIT':
            self.sock.sendto(blob['CHData'], ('127.0.0.1', 56017))
            if self.counter % 100 == 0:
                sys.stdout.write('.')
                sys.stdout.flush()
            self.counter += 1
        return blob


pipe = kp.Pipeline()
pipe.attach(
    kp.io.CHPump,
    host='localhost',
    port=5553,
    tags='IO_MONIT',
    timeout=60 * 60 * 24 * 7,
    max_queue=1000,
    timeit=True
)
pipe.attach(UDPForwarder)
pipe.drain()



# Filename: topology.py
# -*- coding: utf-8 -*-
# pylint: disable=locally-disabled
"""
Topology related modules.

"""

import numpy as np

from km3pipe import Module


class TriggeredDUs(Module):
    """Check for triggered DUs."""
    def process(self, blob):
        triggered_hits = blob['Hits'].triggered_rows.to_dataframe()
        self.calibration.apply(triggered_hits)
        dus = np.unique(triggered_hits['du'])
        n_dus = len(dus)

        blob['TriggeredDUs'] = dus

        if n_dus > 1:
            blob['multiple_du_event'] = True
        else:
            blob['single_du_event'] = True

        return blob



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ts=4 sw=4 et
"""
=====================================================
Check consistency of monitoring channel and TDC times
=====================================================

This script picks a monitoring channel packet with no HRV flags for any PMT
and tries to match one of the following 50 summaryslices using the PMT rates.
The index of the summaryslice, the rates-diff and the time offsets are logged
in a CSV file.

.. code-block:: bash

    Usage:
        tmch_sum_offsets.py DOM_ID [-n N_TIMESLICES]
        tmch_sum_offsets.py (-h | --help)

    Options:
        DOM_ID           The DOM ID.
        -n N_TIMESLICES  The number of timeslices to investigate [default: 50].
        -h --help        Show this screen.

"""

import io
import os
import km3pipe as kp
from km3pipe.io.daq import TMCHData, DAQPreamble, DAQSummaryslice
import numpy as np

__author__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
VERSION = "1.0"

log = kp.logger.get_logger("tmch_sum_offsets")


class MonitoringChannelPicker(kp.Module):
    """Picks a monitoring channel packet with no HRV for a given DOM"""
    def configure(self):
        self.dom_id = self.require("dom_id")

    def process(self, blob):
        if not self.services['SummarysliceFound']():
            return blob
        tag = str(blob['CHPrefix'].tag)
        if tag == 'IO_MONIT':
            tmch_data = TMCHData(io.BytesIO(blob['CHData']))
            dom_id = tmch_data.dom_id
            if dom_id != self.dom_id or tmch_data.hrvbmp > 0:
                log.info("Skipping TMCH packet due to HRV flags")
                return blob
            blob['Candidate'] = tmch_data
            print("Next IO_MONIT candidate picked.")
        return blob


class SummarysliceMatcher(kp.Module):
    def configure(self):
        self.dom_id = self.require("dom_id")
        self.n_timeslices = self.get("n_timeslices", default=50)
        self.expose(self.io_sum_found, 'SummarysliceFound')
        self._candidate = None
        self._rates_io_monit = None
        self._reset()

        filename = "tmch_sum_offsets_{}.csv".format(self.dom_id)
        self.fobj = self._get_file_handler(filename)

    def _reset(self):
        self._io_sum_found = True
        self._diff = []
        self._summaries = []
        self.i = 0

    def _get_file_handler(self, filename):
        if not os.path.exists(filename):
            fobj = open(filename, 'w')
            fobj.write(
                "run dom_id nearest_idx diff "
                "tmch_timestamp tmch_ns "
                "summary_timestamp summary_ns\n"
            )
        else:
            fobj = open(filename, 'a')
        return fobj

    def process(self, blob):
        if 'Candidate' in blob:
            print("Searching for new IO_SUM candidate")
            self._io_sum_found = False
            self._candidate = blob['Candidate']
            self._rates_io_monit = np.array(self._candidate.pmt_rates)

        if self._candidate is None:
            return blob

        tag = str(blob['CHPrefix'].tag)
        if tag == 'IO_SUM':
            data = io.BytesIO(blob['CHData'])
            preamble = DAQPreamble(file_obj=data)    # noqa
            summary = DAQSummaryslice(file_obj=data)
            try:
                rates = np.array(summary.summary_frames[self.dom_id])
            except KeyError:
                print("No DOM data in summaryslice, skipping...")
                return blob
            self.i += 1
            diff = np.sum(np.abs(self._rates_io_monit) - np.abs(rates))
            self._diff.append(diff)
            self._summaries.append(summary)

            if self.i > self.n_timeslices:
                print("Trying to match a summaryslice.")
                idx_nearest = (np.abs(self._diff)).argmin()
                summary = self._summaries[idx_nearest]
                print(
                    "min", min(self._diff), "max", max(self._diff), "nearest",
                    self._diff[idx_nearest]
                )
                print(self._rates_io_monit)
                print(summary.summary_frames[self.dom_id])
                print(
                    "Time of IO_MONIT:", self._candidate.utc_seconds,
                    self._candidate.nanoseconds
                )
                print(
                    "Time of IO_SUM:", summary.header.time_stamp,
                    summary.header.ticks * 16
                )
                self.fobj.write(
                    "{} {} {} {} {} {} {} {}\n".format(
                        self._candidate.run, self.dom_id, idx_nearest,
                        self._diff[idx_nearest], self._candidate.utc_seconds,
                        self._candidate.nanoseconds, summary.header.time_stamp,
                        summary.header.ticks * 16
                    )
                )
                self._reset()
                self.fobj.flush()
        return blob

    def io_sum_found(self):
        """Service which tells if we are ready for the next TMCH packet"""
        return self._io_sum_found

    def finish(self):
        self.fobj.close()


def main():
    from docopt import docopt
    args = docopt(__doc__)

    dom_id = int(args['DOM_ID'])

    pipe = kp.Pipeline(timeit=True)
    pipe.attach(
        kp.io.CHPump,
        host='127.0.0.1',
        port=5553,
        tags='IO_SUM, IO_MONIT',
        timeout=60 * 60 * 24 * 7,
        max_queue=1000
    )
    pipe.attach(MonitoringChannelPicker, dom_id=dom_id)
    pipe.attach(
        SummarysliceMatcher, dom_id=dom_id, n_timeslices=int(args['-n'])
    )
    pipe.drain()


if __name__ == "__main__":
    main()



# Filename: test_stats.py
# pylint: disable=locally-disabled,C0111,R0904,C0103

import numpy as np
from numpy.random import RandomState
from pandas import DataFrame
import pytest

from km3pipe.testing import TestCase
from km3pipe.stats import mad, mad_std, drop_zero_variance, perc, bincenters

__author__ = ["Tamas Gal", "Moritz Lotze"]
__copyright__ = "Copyright 2016, KM3Pipe devs and the KM3NeT collaboration."
__credits__ = ["Thomas Heid"]
__license__ = "MIT"
__maintainer__ = ["Tamas Gal", "Moritz Lotze"]
__email__ = "tgal@km3net.de"
__status__ = "Development"


class TestMAD(TestCase):
    def test_wiki(self):
        arr = np.array([1, 1, 2, 2, 4, 6, 9])
        assert np.allclose(mad(arr), 1)

    def test_normal(self):
        np.random.seed(42)
        sample = np.random.normal(0, 1, 1000)
        assert np.allclose(mad_std(sample), 1, atol=.1)


class TestPerc(TestCase):
    def test_settings(self):
        arr = np.array([
            -2.394,
            0.293,
            0.371,
            0.384,
            1.246,
        ])
        assert np.allclose(perc(arr), [-2.1253, 1.1598])
        assert np.allclose(
            perc(arr, interpolation='nearest'), [arr[0], arr[-1]]
        )


class TestVariance(TestCase):
    def test_easy(self):
        df = DataFrame({'a': [1, 2, 3], 'b': [1, 1, 1], 'c': [4, 4, 3]})
        df2 = drop_zero_variance(df)
        assert df.shape == (3, 3)
        assert df2.shape == (3, 2)
        assert 'b' not in df2.columns
        assert 'b' in df.columns


class TestBins(TestCase):
    def test_binlims(self):
        bins = np.linspace(0, 20, 21)
        assert bincenters(bins).shape[0] == bins.shape[0] - 1



# Filename: plot.py
# pylint: disable=C0103
# pragma: no cover
"""
Common Plotting utils.
"""
try:
    import _tkinter    # noqa
except ImportError:
    import matplotlib
    matplotlib.use('agg')
import matplotlib.pyplot as plt
import numpy as np
try:
    import seaborn as sns
    HAS_SEABORN = True
except ImportError:
    HAS_SEABORN = False

from km3pipe.stats import bincenters

__author__ = "Moritz Lotze"
__copyright__ = "Copyright 2018, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "BSD-3"
__maintainer__ = "Moritz Lotze"
__email__ = "mlotze@km3net.de"
__status__ = "Development"


def hexbin(x, y, color="purple", **kwargs):
    """Seaborn-compatible hexbin plot.

    See also: http://seaborn.pydata.org/tutorial/axis_grids.html#mapping-custom-functions-onto-the-grid
    """
    if HAS_SEABORN:
        cmap = sns.light_palette(color, as_cmap=True)
    else:
        cmap = "Purples"
    plt.hexbin(x, y, cmap=cmap, **kwargs)


def get_ax(ax=None):
    """Grab last ax if none specified, or pass through."""
    if ax is None:
        ax = plt.gca()
    return ax


def diag(ax=None, linecolor='0.0', linestyle='--', **kwargs):
    """Plot the diagonal."""
    ax = get_ax(ax)
    xy_min = np.min((ax.get_xlim(), ax.get_ylim()))
    xy_max = np.max((ax.get_ylim(), ax.get_xlim()))
    return ax.plot([xy_min, xy_max], [xy_min, xy_max],
                   ls=linestyle,
                   c=linecolor,
                   **kwargs)


def automeshgrid(
    x, y, step=0.02, xstep=None, ystep=None, pad=0.5, xpad=None, ypad=None
):
    """Make a meshgrid, inferred from data."""
    if xpad is None:
        xpad = pad
    if xstep is None:
        xstep = step
    if ypad is None:
        ypad = pad
    if ystep is None:
        ystep = step
    xmin = x.min() - xpad
    xmax = x.max() + xpad
    ymin = y.min() - ypad
    ymax = y.max() + ypad
    return meshgrid(xmin, xmax, step, ymin, ymax, ystep)


def meshgrid(x_min, x_max, x_step, y_min=None, y_max=None, y_step=None):
    """Make a meshgrid, e.g. when plotting decision surfaces."""
    if y_min is None:
        y_min = x_min
    if y_max is None:
        y_max = x_max
    if y_step is None:
        y_step = x_step
    xx, yy = np.meshgrid(
        np.arange(x_min, x_max, x_step), np.arange(y_min, y_max, y_step)
    )
    return xx, yy


def prebinned_hist(counts, binlims, ax=None, *args, **kwargs):
    """Plot a histogram with counts, binlims already given.

    Example
    =======
    >>> gaus = np.random.normal(size=100)
    >>> counts, binlims = np.histogram(gaus, bins='auto')
    >>> prebinned_hist(countsl binlims)
    """
    ax = get_ax(ax)
    x = bincenters(binlims)
    weights = counts
    return ax.hist(x, bins=binlims, weights=weights, *args, **kwargs)


def joint_hex(x, y, **kwargs):
    """Seaborn Joint Hexplot with marginal KDE + hists."""
    return sns.jointplot(
        x, y, kind='hex', stat_func=None, marginal_kws={'kde': True}, **kwargs
    )


def plot_convexhull(xy, ax=None, plot_points=True):
    from scipy.spatial import ConvexHull
    ch = ConvexHull(xy)
    ax = get_ax(ax)
    if plot_points:
        ax.plot(xy[:, 0], xy[:, 1], 'o')
    for simplex in ch.simplices:
        ax.plot(xy[simplex, 0], xy[simplex, 1], 'k-')
    return ax



# Filename: db.py
# pylint: disable=locally-disabled
"""
Database utilities.

"""
from datetime import datetime
import functools
import numbers
import ssl
import io
import os
import json
import re
import pytz
import sys
import numpy as np
from collections import defaultdict, OrderedDict, namedtuple
from inspect import Signature, Parameter
from http.cookiejar import CookieJar
from urllib.parse import urlencode, unquote
from urllib.request import (
    Request, build_opener, urlopen, HTTPCookieProcessor, HTTPHandler
)
from urllib.error import URLError, HTTPError
from io import StringIO
from http.client import IncompleteRead

from .tools import cprint
from .time import Timer
from .logger import get_logger

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

log = get_logger(__name__)    # pylint: disable=C0103

UTC_TZ = pytz.timezone('UTC')

# Ignore invalid certificate error
ssl._create_default_https_context = ssl._create_unverified_context

BASE_URL = 'https://km3netdbweb.in2p3.fr'


def we_are_in_lyon():
    """Check if we are on a Lyon machine"""
    import socket
    try:
        hostname = socket.gethostname()
        ip = socket.gethostbyname(hostname)
    except socket.gaierror:
        return False
    return ip.startswith("134.158.")


def we_are_on_jupyterhub():
    """Check if we are on the JupyterHub machine"""
    import socket
    try:
        hostname = socket.gethostname()
        ip = socket.gethostbyname(hostname)
    except socket.gaierror:
        return False
    return ip == socket.gethostbyname("jupyter.km3net.de")


def we_are_on_km3net_gitlab_ci():
    """Check if we are on a GitLab CI runner server"""
    import urllib.request
    external_ip = urllib.request.urlopen('https://ident.me').read().decode('utf8')
    return external_ip == '131.188.161.155'


def read_csv(text, sep="\t"):
    """Create a DataFrame from CSV text"""
    import pandas as pd    # no top level load to make a faster import of db
    return pd.read_csv(StringIO(text), sep="\t")


def make_empty_dataset():
    """Create an empty dataset"""
    import pandas as pd    # no top level load to make a faster import of db
    return pd.DataFrame()


class DBManager(object):
    """A wrapper for the KM3NeT Web DB"""
    def __init__(
        self, username=None, password=None, url=None, temporary=False
    ):
        "Create database connection"
        self._cookies = []
        self._parameters = None
        self._doms = None
        self._detectors = None
        self._opener = None
        self._temporary = temporary

        self.log = get_logger(self.__class__.__name__)

        from .config import Config
        config = Config()

        if url is not None:
            self._db_url = url
        else:
            self._db_url = config.db_url or BASE_URL

        self._login_url = self._db_url + '/home.htm'

        if not temporary:
            if we_are_in_lyon():
                self.restore_session(
                    "sid=_kmcprod_134.158_lyo7783844001343100343mcprod1223user"
                )
                return
            if we_are_on_jupyterhub():
                self.restore_session(
                    "sid=_jupyter-km3net_131.188.161.143_"
                    "d9fe89a1568a49a5ac03bdf15d93d799"
                )
                return
            if we_are_on_km3net_gitlab_ci():
                self.restore_session(
                    "sid=_gitlab-km3net_131.188.161.155_"
                    "f835d56ca6d946efb38324d59e040761"
                )
                return

        if username is not None and password is not None:
            self.login(username, password)
        elif config.db_session_cookie not in (None, ''):
            self.restore_session(config.db_session_cookie)
        elif all(v in os.environ
                 for v in ['KM3NET_DB_USERNAME', 'KM3NET_DB_PASSWORD']):
            login_ok = self.login(
                os.environ['KM3NET_DB_USERNAME'],
                os.environ['KM3NET_DB_PASSWORD']
            )
            if not login_ok:
                self.log.critical("Login failed with credentail from ENV")
                sys.exit(1)
        else:
            username, password = config.db_credentials
            login_ok = self.login(username, password)
            if login_ok and not self._temporary and \
               input("Request permanent session? (y/n)") in 'yY':
                self.request_permanent_session(username, password)

    def datalog(self, parameter, run, maxrun=None, det_id='D_ARCA001'):
        "Retrieve datalogs for given parameter, run(s) and detector"
        parameter = parameter.lower()
        if maxrun is None:
            maxrun = run
        with Timer('Database lookup'):
            return self._datalog(parameter, run, maxrun, det_id)

    def _datalog(self, parameter, run, maxrun, det_id):
        "Extract data from database"
        values = {
            'parameter_name': parameter,
            'minrun': run,
            'maxrun': maxrun,
            'detid': det_id,
        }
        data = urlencode(values)
        content = self._get_content('streamds/datalognumbers.txt?' + data)
        if content.startswith('ERROR'):
            self.log.error(content)
            return None
        try:
            dataframe = read_csv(content)
        except ValueError:
            self.log.warning(
                "Empty dataset"
            )    # ...probably. Waiting for more info
            return make_empty_dataset()
        else:
            add_datetime(dataframe)
            try:
                self._add_converted_units(dataframe, parameter)
            except KeyError:
                self.log.warning(
                    "Could not add converted units for {0}".format(parameter)
                )
            return dataframe

    def run_table(self, det_id='D_ARCA001'):
        url = 'streamds/runs.txt?detid={0}'.format(det_id)
        content = self._get_content(url)
        try:
            df = read_csv(content)
        except ValueError:
            self.log.warning("Empty dataset")
            return None
        else:
            timestamp_column = 'UNIXSTARTTIME'
            add_datetime(df, timestamp_column)
            df['RUNDURATION'] = -df[timestamp_column].diff(periods=-1) / 1000
            return df

    def _add_converted_units(self, dataframe, parameter, key='VALUE'):
        """Add an additional DATA_VALUE column with converted VALUEs"""
        convert_unit = self.parameters.get_converter(parameter)
        try:
            self.log.debug("Adding unit converted DATA_VALUE to the data")
            dataframe[key] = dataframe['DATA_VALUE'].apply(convert_unit)
        except KeyError:
            self.log.warning("Missing 'VALUE': no unit conversion.")
        else:
            dataframe.unit = self.parameters.unit(parameter)

    @property
    def detectors(self):
        if self._detectors is None:
            self._detectors = self._get_detectors()
        return self._detectors

    def _get_detectors(self):
        content = self._get_content('streamds/detectors.txt')
        try:
            dataframe = read_csv(content)
        except ValueError:
            self.log.warning("Empty dataset")
            return make_empty_dataset()
        else:
            return dataframe

    def get_det_id(self, det_oid):
        """Convert detector string representation (OID) to serialnumber"""
        try:
            return self.detectors[self.detectors.OID == det_oid
                                  ].SERIALNUMBER.iloc[0]
        except IndexError:
            self.log.critical("No det ID found for OID '{}'".format(det_oid))
            return None

    def get_det_oid(self, det_id):
        """Convert detector serialnumber to string representation (OID)"""
        try:
            return self.detectors[self.detectors.SERIALNUMBER == det_id
                                  ].OID.iloc[0]
        except IndexError:
            self.log.critical("No OID found for det ID '{}'".format(det_id))
            return None

    def to_det_id(self, det_id_or_det_oid):
        """Convert det ID or OID to det ID"""
        try:
            int(det_id_or_det_oid)
        except ValueError:
            return self.get_det_id(det_id_or_det_oid)
        else:
            return det_id_or_det_oid

    def to_det_oid(self, det_id_or_det_oid):
        """Convert det OID or ID to det OID"""
        try:
            int(det_id_or_det_oid)
        except ValueError:
            return det_id_or_det_oid
        else:
            return self.get_det_oid(det_id_or_det_oid)

    @property
    def parameters(self):
        "Return the parameters container for quick access to their details"
        if self._parameters is None:
            self._load_parameters()
        return self._parameters

    def _load_parameters(self):
        "Retrieve a list of available parameters from the database"
        parameters = self._get_json('allparam/s')
        data = {}
        for parameter in parameters:    # There is a case-chaos in the DB
            data[parameter['Name'].lower()] = parameter
        self._parameters = ParametersContainer(data)

    def trigger_setup(self, runsetup_oid):
        "Retrieve the trigger setup for a given runsetup OID"
        r = self._get_content(
            "jsonds/rslite/s?rs_oid={}&upifilter=1.1.2.2.3/*".
            format(runsetup_oid)
        )
        data = json.loads(r)['Data']
        if not data:
            self.log.error("Empty dataset.")
            return
        raw_setup = data[0]
        det_id = raw_setup['DetID']
        name = raw_setup['Name']
        description = raw_setup['Desc']

        _optical_df = raw_setup['ConfGroups'][0]
        optical_df = {'Name': _optical_df['Name'], 'Desc': _optical_df['Desc']}
        for param in _optical_df['Params']:
            pname = self.parameters.oid2name(param['OID']).replace('DAQ_', '')
            try:
                dtype = float if '.' in param['Val'] else int
                val = dtype(param['Val'])
            except ValueError:
                val = param['Val']
            optical_df[pname] = val

        if len(raw_setup['ConfGroups']) > 1:
            _acoustic_df = raw_setup['ConfGroups'][1]
            acoustic_df = {
                'Name': _acoustic_df['Name'],
                'Desc': _acoustic_df['Desc']
            }
            for param in _acoustic_df['Params']:
                pname = self.parameters.oid2name(param['OID']
                                                 ).replace('DAQ_', '')
                try:
                    dtype = float if '.' in param['Val'] else int
                    val = dtype(param['Val'])
                except ValueError:
                    val = param['Val']
                acoustic_df[pname] = val
        else:
            acoustic_df = {'Not available': None}

        return TriggerSetup(
            runsetup_oid, name, det_id, description, optical_df, acoustic_df
        )

    def detx(self, det_id, t0set=None, calibration=None):
        """Retrieve the detector file for given detector id

        If t0set is given, append the calibration data.
        """
        url = 'detx/{0}?'.format(det_id)    # '?' since it's ignored if no args
        if t0set is not None:
            url += '&t0set=' + t0set
        if calibration is not None:
            url += '&calibrid=' + calibration

        detx = self._get_content(url)
        return detx

    def detx_for_run(self, det_id, run):
        """Retrieve the calibrate detector file for given run"""
        run_table = self.run_table(det_id)
        try:
            run_info = run_table[run_table.RUN == run].iloc[0]
        except IndexError:
            self.log.error(
                "Run {} not found for detector {}".format(run, det_id)
            )

        tcal = run_info.T0_CALIBSETID
        if str(tcal) == 'nan':
            self.log.warning(
                "No time calibration found for run {} (detector {})".format(
                    run, det_id
                )
            )
            tcal = 0

        try:
            pcal = int(run_info.POS_CALIBSETID)
        except ValueError:
            self.log.warning(
                "No position calibration found for run {} (detector {})".
                format(run, det_id)
            )
            pcal = 0

        try:
            rcal = int(run_info.ROT_CALIBSETID)
        except ValueError:
            self.log.warning(
                "No rotation calibration found for run {} (detector {})".
                format(run, det_id)
            )
            rcal = 0

        url = 'detx/{det_id}?tcal={tcal}&pcal={pcal}&rcal={rcal}'.format(
            det_id=det_id, tcal=tcal, pcal=pcal, rcal=rcal
        )

        detx = self._get_content(url)
        return detx

    def ahrs(self, run, maxrun=None, clbupi=None, det_id='D_ARCA001'):
        "Retrieve AHRS values for given run(s) (optionally CLBs) and detector"
        if maxrun is None:
            maxrun = run
        with Timer('Database lookup'):
            return self._ahrs(run, maxrun, clbupi, det_id)

    def _ahrs(self, run, maxrun, clbupi, det_id):
        values = {
            'minrun': run,
            'maxrun': maxrun,
            'detid': det_id,
        }
        if clbupi is not None:
            values['clbupi'] = clbupi
        data = urlencode(values)
        content = self._get_content('streamds/ahrs.txt?' + data)
        if content.startswith('ERROR'):
            self.log.error(content)
            return None
        try:
            dataframe = read_csv(content)
        except ValueError:
            self.log.warning(
                "Empty dataset"
            )    # ...probably. Waiting for more info
            return make_empty_dataset()
        else:
            add_datetime(dataframe)
            return dataframe

    def _get_json(self, url):
        "Get JSON-type content"
        content = self._get_content('jsonds/' + url)
        try:
            json_content = json.loads(content.decode())
        except AttributeError:
            json_content = json.loads(content)
        if json_content.get('Comment') is not None:
            self.log.warning(json_content['Comment'])
        if json_content['Result'] != 'OK':
            self.log.critical("Error from DB: %s", json_content.get('Data'))
            raise ValueError("Error while retrieving the parameter list.")
        return json_content['Data']

    def _get_content(self, url):
        "Get HTML content"
        target_url = self._db_url + '/' + unquote(url)    # .encode('utf-8'))
        self.log.debug("Opening '{0}'".format(target_url))
        try:
            f = self.opener.open(target_url)
        except HTTPError as e:
            self.log.error(
                "HTTP error, your session may be expired.\n"
                "Original HTTP error: {}\n"
                "Target URL: {}".format(e, target_url)
            )
            if input("Request new permanent session and retry? (y/n)") in 'yY':
                self.request_permanent_session()
                return self._get_content(url)
            else:
                return None
        self.log.debug("Accessing '{0}'".format(target_url))
        try:
            content = f.read()
        except IncompleteRead as icread:
            self.log.critical(
                "Incomplete data received from the DB, " +
                "the data could be corrupted."
            )
            content = icread.partial
        self.log.debug("Got {0} bytes of data.".format(len(content)))
        return content.decode('utf-8')

    @property
    def opener(self):
        "A reusable connection manager"
        if self._opener is None:
            self.log.debug("Creating connection handler")
            opener = build_opener()
            if self._cookies:
                self.log.debug("Appending cookies")
            else:
                self.log.debug("No cookies to append")
            for cookie in self._cookies:
                cookie_str = cookie.name + '=' + cookie.value
                opener.addheaders.append(('Cookie', cookie_str))
            self._opener = opener
        else:
            self.log.debug("Reusing connection manager")
        return self._opener

    def request_sid_cookie(self, username, password):
        """Request cookie for permanent session token."""
        self.log.debug("Requesting SID cookie")
        target_url = self._login_url + '?usr={0}&pwd={1}&persist=y'.format(
            username, password
        )
        cookie = urlopen(target_url).read()
        return cookie

    def restore_session(self, cookie):
        """Establish databse connection using permanent session cookie"""
        self.log.debug("Restoring session from cookie: {}".format(cookie))
        opener = build_opener()
        opener.addheaders.append(('Cookie', cookie))
        self._opener = opener

    def request_permanent_session(self, username=None, password=None):
        self.log.debug("Requesting permanent session")

        from .config import Config
        config = Config()

        if username is None and password is None:
            self.log.debug("Checking configuration file for DB credentials")
            username, password = config.db_credentials
        cookie = self.request_sid_cookie(username, password)
        cookie_str = str(cookie, 'utf-8')    # Python 3
        self.log.debug("Session cookie: {0}".format(cookie_str))
        self.log.debug("Storing cookie in configuration file")
        config.set('DB', 'session_cookie', cookie_str)
        # self._cookies = [cookie]
        self.restore_session(cookie)

    def login(self, username, password):
        "Login to the database and store cookies for upcoming requests."
        self.log.debug("Logging in to the DB")
        opener = self._build_opener()
        values = {'usr': username, 'pwd': password}
        req = self._make_request(self._login_url, values)
        try:
            self.log.debug("Sending login request")
            f = opener.open(req)
        except URLError as e:
            self.log.error(
                "Failed to connect to the database -> probably down!"
            )
            self.log.error("Error from database server:\n    {0}".format(e))
            return False
        html = f.read()
        failed_auth_message = 'Bad username or password'
        if failed_auth_message in str(html):
            self.log.error(failed_auth_message)
            return False
        return True

    def _build_opener(self):
        self.log.debug("Building opener.")
        cj = CookieJar()
        self._cookies = cj
        opener = build_opener(HTTPCookieProcessor(cj), HTTPHandler())
        return opener

    def _make_request(self, url, values):
        data = urlencode(values)
        return Request(url, data.encode('utf-8'))

    def _post(self, url, data):
        pass


def add_datetime(dataframe, timestamp_key='UNIXTIME', scale_factor=1e3):
    """Add an additional DATETIME column with standar datetime format.

    This currently manipulates the incoming DataFrame!
    """
    def convert_data(timestamp):
        return datetime.fromtimestamp(float(timestamp) / scale_factor, UTC_TZ)

    try:
        log.debug("Adding DATETIME column to the data")
        converted = dataframe[timestamp_key].apply(convert_data)
        dataframe['DATETIME'] = converted
    except KeyError:
        log.warning("Could not add DATETIME column")


class StreamDS(object):
    """Access to the streamds data stored in the KM3NeT database."""
    def __init__(
        self, username=None, password=None, url=None, temporary=False
    ):
        self._db = DBManager(username, password, url, temporary)
        self._stream_df = None
        self._streams = None

        self._update_streams()

    def _update_streams(self):
        """Update the list of available straems"""
        content = self._db._get_content("streamds")
        self._stream_df = read_csv(content).sort_values("STREAM")
        self._streams = None
        for stream in self.streams:
            setattr(self, stream, self.__getattr__(stream))

    def __getattr__(self, attr):
        """Magic getter which optionally populates the function signatures"""
        if attr in self.streams:
            stream = attr
        else:
            raise AttributeError

        def func(**kwargs):
            return self.get(stream, **kwargs)

        func.__doc__ = self._stream_parameter(stream, "DESCRIPTION")

        sig_dict = OrderedDict()
        for sel in self.mandatory_selectors(stream):
            if sel == '-':
                continue
            sig_dict[Parameter(sel, Parameter.POSITIONAL_OR_KEYWORD)] = None
        for sel in self.optional_selectors(stream):
            if sel == '-':
                continue
            sig_dict[Parameter(sel, Parameter.KEYWORD_ONLY)] = None
        func.__signature__ = Signature(parameters=sig_dict)

        return func

    @property
    def streams(self):
        """A list of available streams"""
        if self._streams is None:
            self._streams = list(self._stream_df["STREAM"].values)
        return self._streams

    def _stream_parameter(self, stream, parameter):
        data = self._stream_df[self._stream_df.STREAM == stream]
        if 'SELECTORS' in parameter:
            return list(data[parameter].values[0].split(','))
        else:
            return data[parameter].values[0]

    def mandatory_selectors(self, stream):
        """A list of mandatory selectors for a given stream"""
        return self._stream_parameter(stream, 'MANDATORY_SELECTORS')

    def optional_selectors(self, stream):
        """A list of optional selectors for a given stream"""
        return self._stream_parameter(stream, 'OPTIONAL_SELECTORS')

    def help(self, stream):
        """Show the help for a given stream."""
        if stream not in self.streams:
            log.error("Stream '{}' not found in the database.".format(stream))
        params = self._stream_df[self._stream_df['STREAM'] == stream].values[0]
        self._print_stream_parameters(params)

    def print_streams(self):
        """Print a coloured list of streams and its parameters"""
        for row in self._stream_df.itertuples():
            self._print_stream_parameters(row[1:])

    def _print_stream_parameters(self, values):
        """Print a coloured help for a given tuple of stream parameters."""
        cprint("{0}".format(*values), "magenta", attrs=["bold"])
        print("{4}".format(*values))
        cprint("  available formats:   {1}".format(*values), "blue")
        cprint("  mandatory selectors: {2}".format(*values), "red")
        cprint("  optional selectors:  {3}".format(*values), "green")
        print()

    def get(self, stream, fmt='txt', **kwargs):
        """Get the data for a given stream manually"""
        sel = ''.join(["&{0}={1}".format(k, v) for (k, v) in kwargs.items()])
        url = "streamds/{0}.{1}?{2}".format(stream, fmt, sel[1:])
        data = self._db._get_content(url)
        if not data:
            log.error("No data found at URL '%s'." % url)
            return
        if (data.startswith("ERROR")):
            log.error(data)
            return
        if fmt == "txt":
            return read_csv(data)
        return data


class ParametersContainer(object):
    """Provides easy access to parameters"""
    def __init__(self, parameters):
        self._parameters = parameters
        self._converters = {}
        self._oid_lookup = defaultdict(lambda: None)

    @property
    def names(self):
        "A list of parameter names"
        return list(self._parameters.keys())

    def get_parameter(self, parameter):
        "Return a dict for given parameter"
        parameter = self._get_parameter_name(parameter)
        return self._parameters[parameter]

    def get_converter(self, parameter):
        """Generate unit conversion function for given parameter"""
        if parameter not in self._converters:
            param = self.get_parameter(parameter)
            try:
                scale = float(param['Scale'])
            except KeyError:
                scale = 1

            def convert(value):
                # easy_scale = float(param['EasyScale'])
                # easy_scale_multiplier = float(param['EasyScaleMultiplier'])
                return value * scale

            return convert

    def unit(self, parameter):
        "Get the unit for given parameter"
        parameter = self._get_parameter_name(parameter).lower()
        return self._parameters[parameter]['Unit']

    def oid2name(self, oid):
        "Look up the parameter name for a given OID"
        if not self._oid_lookup:
            for name, data in self._parameters.items():
                self._oid_lookup[data['OID']] = data['Name']
        return self._oid_lookup[oid]

    def _get_parameter_name(self, name):
        if name in self.names:
            return name

        aliases = [n for n in self.names if n.endswith(' ' + name)]
        if len(aliases) == 1:
            log.info("Alias found for {0}: {1}".format(name, aliases[0]))
            return aliases[0]

        log.info(
            "Parameter '{0}' not found, trying to find alternative.".
            format(name)
        )
        try:
            # ahrs_g[0] for example should be looked up as ahrs_g
            alternative = re.findall(r'(.*)\[[0-9*]\]', name)[0]
            log.info("Found alternative: '{0}'".format(alternative))
            return alternative
        except IndexError:
            raise KeyError("Could not find alternative for '{0}'".format(name))


class DOM(object):
    """Represents a DOM"""
    def __init__(self, clb_upi, dom_id, dom_upi, du, det_oid, floor):
        self.clb_upi = clb_upi
        self.dom_id = dom_id
        self.dom_upi = dom_upi
        self.du = du
        self.det_oid = det_oid
        self.floor = floor

    @classmethod
    def from_json(cls, json):
        return cls(
            json["CLBUPI"], json["DOMId"], json["DOMUPI"], json["DU"],
            json["DetOID"], json["Floor"]
        )

    @property
    def omkey(self):
        return (self.du, self.floor)

    def __str__(self):
        return "DU{0}-DOM{1}".format(self.du, self.floor)

    def __repr__(self):
        return (
            "{0} - DOM ID: {1}\n"
            "   DOM UPI: {2}\n"
            "   CLB UPI: {3}\n"
            "   DET OID: {4}\n".format(
                self.__str__(), self.dom_id, self.dom_upi, self.clb_upi,
                self.det_oid
            )
        )


class TriggerSetup(object):
    def __init__(
        self, runsetup_oid, name, det_id, description, optical_df, acoustic_df
    ):
        self.runsetup_oid = runsetup_oid
        self.name = name
        self.det_id = det_id
        self.description = description
        self.optical_df = optical_df
        self.acoustic_df = acoustic_df

    def __str__(self):
        text = (
            "Runsetup OID: {}\n"
            "Name: {}\n"
            "Detector ID: {}\n"
            "Description:\n    {}\n\n".format(
                self.runsetup_oid, self.name, self.det_id, self.description
            )
        )
        for df, parameters in zip(["Optical", "Acoustic"],
                                  [self.optical_df, self.acoustic_df]):
            text += "{} Datafilter:\n".format(df)
            for parameter, value in parameters.items():
                text += "  {}: {}\n".format(parameter, value)
            text += "\n"
        return text

    def __repr__(self):
        return str(self)


@functools.lru_cache()
def clbupi2ahrsupi(clb_upi):
    """Generate AHRS UPI from CLB UPI."""
    sds = StreamDS()
    upis = sds.integration(container_upi=clb_upi).CONTENT_UPI.values
    ahrs_upis = [upi for upi in upis if "AHRS" in upi]
    if len(ahrs_upis) > 1:
        log.warning(
            "Multiple AHRS UPIs found for CLB UPI {}. "
            "Using the first entry.".format(clb_upi)
        )
    return ahrs_upis[0]


def show_ahrs_calibration(clb_upi, version='3'):
    """Show AHRS calibration data for given `clb_upi`."""
    db = DBManager()
    ahrs_upi = clbupi2ahrsupi(clb_upi)
    print("AHRS UPI: {}".format(ahrs_upi))
    content = db._get_content("show_product_test.htm?upi={0}&"
                              "testtype=AHRS-CALIBRATION-v{1}&n=1&out=xml"
                              .format(ahrs_upi, version)) \
        .replace('\n', '')

    import xml.etree.ElementTree as ET

    try:
        root = ET.parse(io.StringIO(content)).getroot()
    except ET.ParseError:
        print("No calibration data found")
    else:
        for child in root:
            print("{}: {}".format(child.tag, child.text))
        names = [c.text for c in root.findall(".//Name")]
        values = [[i.text for i in c] for c in root.findall(".//Values")]
        for name, value in zip(names, values):
            print("{}: {}".format(name, value))


class CLBMap(object):
    par_map = {'DETOID': 'det_oid', 'UPI': 'upi', 'DOMID': 'dom_id'}

    def __init__(self, det_oid):
        self.log = get_logger('CLBMap')
        if isinstance(det_oid, numbers.Integral):
            db = DBManager()
            _det_oid = db.get_det_oid(det_oid)
            if _det_oid is not None:
                det_oid = _det_oid
        self.det_oid = det_oid
        sds = StreamDS()
        self._data = sds.clbmap(detoid=det_oid)
        self._by = {}

    def __len__(self):
        return len(self._data)

    @property
    def upis(self):
        """A dict of CLBs with UPI as key"""
        parameter = 'UPI'
        if parameter not in self._by:
            self._populate(by=parameter)
        return self._by[parameter]

    @property
    def dom_ids(self):
        """A dict of CLBs with DOM ID as key"""
        parameter = 'DOMID'
        if parameter not in self._by:
            self._populate(by=parameter)
        return self._by[parameter]

    @property
    def omkeys(self):
        """A dict of CLBs with the OMKey tuple (DU, floor) as key"""
        parameter = 'omkey'
        if parameter not in self._by:
            self._by[parameter] = {}
            for clb in self.upis.values():
                omkey = (clb.du, clb.floor)
                self._by[parameter][omkey] = clb
            pass
        return self._by[parameter]

    def base(self, du):
        """Return the base CLB for a given DU"""
        parameter = 'base'
        if parameter not in self._by:
            self._by[parameter] = {}
            for clb in self.upis.values():
                if clb.floor == 0:
                    self._by[parameter][clb.du] = clb
        return self._by[parameter][du]

    def _populate(self, by):
        data = {}
        for _, row in self._data.iterrows():
            data[row[by]] = CLB(
                det_oid=row['DETOID'],
                floor=row['FLOORID'],
                du=row['DUID'],
                serial_number=row['SERIALNUMBER'],
                upi=row['UPI'],
                dom_id=row['DOMID']
            )
        self._by[by] = data


CLB = namedtuple(
    'CLB', ['det_oid', 'floor', 'du', 'serial_number', 'upi', 'dom_id']
)



# Filename: config.py
# pylint: disable=locally-disabled
"""
Tools for global configuration.

"""
import os
import pytz
from configparser import ConfigParser, Error, NoOptionError, NoSectionError
import getpass

from .logger import get_logger

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal and Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"

log = get_logger(__name__)    # pylint: disable=C0103

CONFIG_PATH = os.path.expanduser('~/.km3net')


class Config(object):
    def __init__(self, config_path=CONFIG_PATH):
        """Configuration manager for KM3NeT stuff"""
        self.config = ConfigParser()
        self._time_zone = None
        self._config_path = config_path

        if config_path is not None:
            self._init_from_path(config_path)

    def _init_from_path(self, path):
        if not os.path.exists(path):
            log.info("No configuration found at '{0}'".format(path))
            return
        self._read_from_path(path)

    def _read_from_path(self, path):
        """Read configuration from file path"""
        with open(path) as config_file:
            self._read_from_file(config_file)

    def _read_from_file(self, file_obj):
        self.config.read_file(file_obj)

    def set(self, section, key, value):
        if section not in self.config.sections():
            self.config.add_section(section)
        self.config.set(section, key, value)
        with open(self._config_path, 'w') as f:
            self.config.write(f)

    def get(self, section, key, default=None):
        try:
            value = self.config.get(section, key)
            try:
                return float(value)
            except ValueError:
                return value
        except (NoOptionError, NoSectionError):
            return default

    def create_irods_session(self):
        try:
            from irods.session import iRODSSession
        except ImportError:
            log.error(
                "Please install the iRODS Python client:\n\n"
                "    pip install git+git://github.com/irods/"
                "python-irodsclient.git\n"
            )
            return
        try:
            host = self.config.get('iRODS', 'host')
            port = self.config.get('iRODS', 'port')
            user = self.config.get('iRODS', 'user')
            zone = self.config.get('iRODS', 'zone')
        except Error:
            log.error("iRODS connection details missing from ~/.km3net")
            return

        try:
            password = self.config.get('iRODS', 'password')
        except Error:
            password = input("Please enter your iRODS password: ")

        return iRODSSession(
            host=host, port=port, user=user, password=password, zone=zone
        )

    @property
    def db_credentials(self):
        """Return username and password for the KM3NeT WebDB."""
        try:
            username = self.config.get('DB', 'username')
            password = self.config.get('DB', 'password')
        except Error:
            username = input("Please enter your KM3NeT DB username: ")
            password = getpass.getpass("Password: ")
        return username, password

    @property
    def db_session_cookie(self):
        try:
            return self.config.get('DB', 'session_cookie')
        except Error:
            return None

    @property
    def db_url(self):
        try:
            return self.config.get('DB', 'url')
        except Error:
            return None

    @property
    def slack_token(self):
        """Return slack token for chat bots."""
        try:
            token = self.config.get('Slack', 'token')
        except Error:
            raise ValueError("No Slack token defined in configuration file.")
        else:
            return token

    @property
    def rba_url(self):
        """Return the RainbowAlga URL."""
        try:
            url = self.config.get('RainbowAlga', 'url')
        except Error:
            return None
        else:
            return url

    @property
    def check_for_updates(self):
        try:
            return self.config.getboolean('General', 'check_for_updates')
        except Error:
            return True

    @property
    def time_zone(self):
        if not self._time_zone:
            try:
                time_zone = self.config.get('General', 'time_zone')
            except Error:
                time_zone = 'UTC'
            self._time_zone = pytz.timezone(time_zone)
        return self._time_zone



# Filename: triggersetup.py
"""
Prints the trigger information of a given run setup.

Usage:
    triggersetup RUNSETUP_OID
    triggersetup (-h | --help)
    triggersetup --version

Options:
    RUNSETUP_OID   The run setup identifier (e.g. A02004580)
    -h --help      Show this screen.

"""

import km3pipe as kp

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2017, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


def main():
    from docopt import docopt
    args = docopt(__doc__, version=kp.version)

    db = kp.db.DBManager()
    print(db.trigger_setup(args['RUNSETUP_OID']))



#!/usr/bin/env python3

from ..core import Module, Blob
from ..dataclasses import Table
from .hdf5 import HDF5Header

import km3io
import numpy as np
from collections import defaultdict

USR_MC_TRACKS_KEYS = [b'energy_lost_in_can', b'bx', b'by', b'ichan', b'cc']


class OfflinePump(Module):
    def configure(self):
        self._filename = self.get("filename")

        self._reader = km3io.OfflineReader(self._filename)
        self.header = self._reader.header
        self.blobs = self._blob_generator()

    def process(self, blob=None):
        return next(self.blobs)

    def finish(self):
        self._reader.close()

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.blobs)

    def __getitem__(self, item):
        if not isinstance(item, int):
            raise TypeError("Only integer indices are supported.")
        return Blob({
            'event': self._reader.events[item],
            'header': self.header
        })

    def get_number_of_blobs(self):
        return len(self._reader.events)

    def _blob_generator(self):
        for event in self._reader.events:
            blob = Blob({'event': event, 'header': self.header})
            yield blob


class EventPump(Module):
    def configure(self):
        self._filename = self.require("filename")
        self._split_hits = self.get("split_hits", default=True)
        self.skip_hits = self.get("skip_hits", default=False)
        self.skip_mc_hits = self.get("skip_mc_hits", default=False)
        self.skip_mc_tracks = self.get("skip_mc_tracks", default=False)
        self.skip_header = self.get("skip_header", default=False)

        self._read_usr_data = self.get("read_usr_data", default=False)
        if self._read_usr_data:
            self.log.warning(
                "Reading usr-data will massively decrease the performance."
            )

        self._reader = km3io.OfflineReader(self._filename)

        self.header = None
        self.raw_header = None
        if not self.skip_header:
            if self._reader.header is not None:
                self.header = HDF5Header.from_km3io(self._reader.header)
                self.raw_header = self._generate_raw_header()

        self.blobs = self._blob_generator()

    def process(self, blob=None):
        return next(self.blobs)

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.blobs)

    def __getitem__(self, item):
        if not isinstance(item, int):
            raise TypeError("Only integer indices are supported.")
        return self._parse_event(self._reader.events[item], Blob())

    def get_number_of_blobs(self):
        return len(self._reader.events)

    def _blob_generator(self):
        for event in self._reader.events:
            blob = Blob()

            self._parse_event(event, blob)
            yield blob

    def _parse_event(self, event, blob):
        blob["RawHeader"] = self.raw_header
        blob["Header"] = self.header

        blob['Event'] = event
        if not self.skip_hits and event.n_hits > 0:
            blob['Hits'] = self._parse_hits(event.hits)
        if not self.skip_mc_hits and event.n_mc_hits > 0:
            blob['McHits'] = self._parse_mc_hits(event.mc_hits)
        if not self.skip_mc_tracks and event.n_mc_tracks > 0:
            blob['McTracks'] = self._parse_mc_tracks(event.mc_tracks)
        return blob

    def _generate_raw_header(self):
        header = {}
        raw_header = {"field_names": [], "parameter": [], "field_values": []}
        for n, x in self._reader._fobj['Head'
                                       ]._map_3c_string_2c_string_3e_.items():
            header[n.decode("utf-8")] = x.decode("utf-8").strip()
        for attribute, fields in km3io.definitions.mc_header.items():
            values = header.get(attribute, '').split()
            if not values:
                continue
            raw_header["field_values"].append(" ".join(map(str, values)))
            raw_header["field_names"].append(" ".join(fields))
            raw_header["parameter"].append(attribute)
        return Table(
            raw_header,
            h5loc="/raw_header",
            name="RawHeader",
            h5singleton=True
        )

    def _parse_eventinfo(self, event):
        wgt1, wgt2, wgt3, wgt4 = self._parse_wgts(event.w)
        tab_data = {
            'event_id': event.id,
            'run_id': event.run_id,
            'weight_w1': wgt1,
            'weight_w2': wgt2,
            'weight_w3': wgt3,
            'weight_w4': wgt4,
            'timestamp': event.t_sec,
            'nanoseconds': event.t_ns,
            'mc_time': event.mc_t,
            'trigger_mask': event.trigger_mask,
            'trigger_counter': event.trigger_counter,
            'overlays': event.overlays,
            'det_id': event.det_id,
            'frame_index': event.frame_index,
            'mc_run_id': event.mc_run_id,
        }
        info = Table(tab_data, h5loc='/event_info', name='EventInfo')
        return info

    @staticmethod
    def _parse_wgts(wgt):
        if len(wgt) == 3:
            wgt1, wgt2, wgt3 = wgt
            wgt4 = np.nan
        elif len(wgt) == 4:
            # what the hell is w4?
            wgt1, wgt2, wgt3, wgt4 = wgt
        else:
            wgt1 = wgt2 = wgt3 = wgt4 = np.nan
        return wgt1, wgt2, wgt3, wgt4

    def _parse_usr_to_dct(self, mc_tracks):
        dct = defaultdict(list)
        for k in USR_MC_TRACKS_KEYS:
            dec_key = k.decode('utf_8')
            for i in range(mc_tracks.usr_names.shape[0]):
                value = np.nan
                if k in mc_tracks.usr_names[i]:
                    mask = mc_tracks.usr_names[i] == k
                    value = mc_tracks.usr[i][mask][0]
                dct[dec_key].append(value)
        return dct

    def _parse_mc_tracks(self, mc_tracks):
        dct = {
            'dir_x': mc_tracks.dir_x,
            'dir_y': mc_tracks.dir_y,
            'dir_z': mc_tracks.dir_z,
            'pos_x': mc_tracks.pos_x,
            'pos_y': mc_tracks.pos_y,
            'pos_z': mc_tracks.pos_z,
            'energy': mc_tracks.E,
            'time': mc_tracks.t,
            'type': mc_tracks.type,
            'id': mc_tracks.id,
            'length': mc_tracks.len
        }
        if self._read_usr_data:
            dct.update(self._parse_usr_to_dct(mc_tracks))
        return Table(dct, name='McTracks', h5loc='/mc_tracks', split_h5=True)

    def _parse_mc_hits(self, mc_hits):
        return Table({
            "a": mc_hits.a,
            "origin": mc_hits.origin,
            "pmt_id": mc_hits.pmt_id,
            "time": mc_hits.t
        },
                     h5loc="/mc_hits",
                     split_h5=self._split_hits,
                     name='McHits')

    def _parse_hits(self, hits):
        return Table({
            "channel_id": hits.channel_id,
            "dom_id": hits.dom_id,
            "time": hits.t,
            "tot": hits.tot,
            "triggered": hits.trig
        },
                     h5loc="/hits",
                     split_h5=self._split_hits,
                     name='Hits')



#!/usr/bin/env python
# Filename: totmonitor.py
"""
Display the mean ToT for each TDC channel for given DOM.

Usage:
    totmonitor [options] DOM_ID
    totmonitor (-h | --help)
    totmonitor --version

Options:
    -l LIGIER       Ligier address [default: 127.0.0.1].
    -p PORT         Ligier port [default: 5553].
    -o OPTIMAL_TOT  Target ToT in ns [default: 26.4].
    -t TOLERANCE    Defines the range for valid ToT [default: 0.3].
    -u UPDATE       Update frequency in seconds [default: 10].
    -h --help       Show this screen.
"""

import os
from collections import defaultdict

from km3pipe.dataclasses import Table
from km3pipe.dev import cprint
from km3pipe.core import Module

import numpy as np
from sklearn import mixture

__author__ = "Tamas Gal and Jonas Reubelt"
__copyright__ = "Copyright 2017, the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal and Jonas Reubelt"
__email__ = "tgal@km3net.de"
__status__ = "Development"

GMM = mixture.GaussianMixture(n_components=1)


class TimesliceCreator(Module):
    """Create `TimesliceHitSeries` from raw timeslice hits."""
    def configure(self):
        self.dom_id = self.require("dom_id")

    def process(self, blob):
        hits = blob['TimesliceFrames'][self.dom_id]
        n_hits = len(hits)
        if n_hits == 0:
            return blob
        channel_ids, times, tots = zip(*hits)
        ts_hits = Table({
            'channel_id': np.array(channel_ids),
            'dom_id': np.full(n_hits, self.dom_id),
            'time': np.array(times),
            'tot': np.array(tots),
            'group_id': 0
        })
        blob['TimesliceHits'] = ts_hits
        blob['DOM_ID'] = self.dom_id
        return blob


class MeanTotDisplay(Module):
    def configure(self):
        self.optimal_tot = self.get("optimal_tot") or 26.4
        self.tolerance = self.get("tolerance") or 0.3
        self.update_frequency = self.get("update_frequency") or 10
        self.tots = defaultdict(list)
        self.counter = 0
        self.dom_id = None

    def process(self, blob):
        hits = blob["TimesliceHits"]
        self.dom_id = blob["DOM_ID"]
        for channel in range(31):
            idx = hits.channel_id == channel
            tots = hits.tot[idx]
            self.tots[channel] += list(tots)

        self.counter += 1
        if self.counter % int(self.update_frequency * 10) == 0:
            self.update_display()
            self.tots = defaultdict(list)
            self.counter = 0
        return blob

    def update_display(self):
        os.system('clear')
        self.print_header()
        for channel, tots in self.tots.items():
            if channel % 8 == 0:
                self.print_scale()
            GMM.fit(np.array(tots)[:, np.newaxis])
            mean_tot = GMM.means_[np.argmin(GMM.covariances_)][0]
            # mean_tot = np.median(tots)
            if np.isnan(mean_tot):
                mean_tot = 0
            color = 'green'
            if mean_tot > self.optimal_tot + self.tolerance:
                color = 'red'
            if mean_tot < self.optimal_tot - self.tolerance:
                color = 'blue'
            cprint(
                "Channel {0:02d}: {1:.1f}ns    {2}".format(
                    channel, mean_tot,
                    int(mean_tot) * '|'
                ), color
            )
        self.print_scale()
        self.print_footer()

    def print_header(self):
        print(
            "Mean ToT (average over {0}s) for DOM: {1}".format(
                self.update_frequency, self.dom_id
            )
        )
        print(
            "                     "
            "0         10        20        30        40        50"
        )

    def print_footer(self):
        print(
            "                     "
            "0         10        20        30        40        50"
        )

    def print_scale(self):
        print("                     " + '|----+----' * 5)


def main():
    from docopt import docopt
    import km3pipe as kp
    args = docopt(__doc__)
    pipe = kp.Pipeline()
    pipe.attach(
        kp.io.ch.CHPump,
        host=args['-l'],
        port=int(args['-p']),
        tags='IO_TSL',
        max_queue=100,
        timeout=60 * 60 * 24
    )
    pipe.attach(kp.io.daq.TimesliceParser)
    pipe.attach(TimesliceCreator, dom_id=int(args['DOM_ID']))
    pipe.attach(
        MeanTotDisplay,
        only_if="TimesliceHits",
        optimal_tot=float(args['-o']),
        update_frequency=float(args['-u']),
        tolerance=float(args['-t'])
    )
    pipe.drain()


if __name__ == "__main__":
    main()



#!usr/bin/env python
# -*- coding: utf-8 -*-
# Filename: stats.py
# pylint: disable=C0103
"""
Statistics.
"""
import numpy as np

from .logger import get_logger

__author__ = "Moritz Lotze"
__copyright__ = "Copyright 2017, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Moritz Lotze"
__email__ = "mlotze@km3net.de"
__status__ = "Development"

log = get_logger(__name__)    # pylint: disable=C0103


def mad(v):
    """MAD -- Median absolute deviation. More robust than standard deviation.
    """
    return np.median(np.abs(v - np.median(v)))


def mad_std(v):
    """Robust estimate of standard deviation using the MAD.

    Lifted from astropy.stats."""
    MAD = mad(v)
    return MAD * 1.482602218505602


def drop_zero_variance(df):
    """Remove columns from dataframe with zero variance."""
    return df.copy().loc[:, df.var() != 0].copy()


def perc(arr, p=95, **kwargs):
    """Create symmetric percentiles, with ``p`` coverage."""
    offset = (100 - p) / 2
    return np.percentile(arr, (offset, 100 - offset), **kwargs)


def bincenters(bins):
    """Bincenters, assuming they are all equally spaced."""
    bins = np.atleast_1d(bins)
    return 0.5 * (bins[1:] + bins[:-1])



# -*- coding: utf-8 -*-
"""
==================
DOM hits.
==================

Estimate track/DOM distances using the number of hits per DOM.
"""

# Author: Tamas Gal <tgal@km3net.de>
# License: BSD-3

from collections import defaultdict, Counter

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm

import km3pipe as kp
from km3pipe.dataclasses import Table
from km3pipe.math import pld3
from km3modules.common import StatusBar
import km3pipe.style
km3pipe.style.use("km3pipe")

filename = "data/atmospheric_muons_sample.h5"
cal = kp.calib.Calibration(filename="data/KM3NeT_-00000001_20171212.detx")


def filter_muons(blob):
    """Write all muons from McTracks to Muons."""
    tracks = blob['McTracks']
    muons = tracks[tracks.type == -13]    # PDG particle code
    blob["Muons"] = Table(muons)
    return blob


class DOMHits(kp.Module):
    """Create histogram with n_hits and distance of hit to track."""
    def configure(self):
        self.hit_statistics = defaultdict(list)

    def process(self, blob):
        hits = blob['Hits']
        muons = blob['Muons']

        highest_energetic_muon = Table(muons[np.argmax(muons.energy)])
        muon = highest_energetic_muon

        triggered_hits = hits.triggered_rows

        dom_hits = Counter(triggered_hits.dom_id)
        for dom_id, n_hits in dom_hits.items():
            try:
                distance = pld3(
                    cal.detector.dom_positions[dom_id], muon.pos, muon.dir
                )
            except KeyError:
                self.log.warning("DOM ID %s not found!" % dom_id)
                continue
            self.hit_statistics['n_hits'].append(n_hits)
            self.hit_statistics['distance'].append(distance)
        return blob

    def finish(self):
        df = pd.DataFrame(self.hit_statistics)
        print(df)
        sdf = df[(df['distance'] < 200) & (df['n_hits'] < 50)]
        bins = (int(max(sdf['distance'])) - 1, int(max(sdf['n_hits']) - 1))
        plt.hist2d(
            sdf['distance'],
            sdf['n_hits'],
            cmap='plasma',
            bins=bins,
            norm=LogNorm()
        )
        plt.xlabel('Distance between hit and muon track [m]')
        plt.ylabel('Number of hits on DOM')
        plt.show()


pipe = kp.Pipeline()
pipe.attach(kp.io.HDF5Pump, filename=filename)
pipe.attach(StatusBar, every=100)
pipe.attach(filter_muons)
pipe.attach(DOMHits)
pipe.drain()



# Filename: test_controlhost.py
# pylint: disable=locally-disabled,C0111,R0904,R0201,C0103,W0612
"""
Unit tests for the controlhost module.

"""

from km3pipe.controlhost import Tag, Message, Prefix

from km3pipe.testing import TestCase

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class TestTag(TestCase):
    def test_empty_tag_has_correct_length(self):
        tag = Tag()
        self.assertEqual(Tag.SIZE, len(tag))

    def test_tag_has_correct_length(self):
        for tag_name in (b'foo', b'bar', b'baz', b'1'):
            tag = Tag(tag_name)
            self.assertEqual(Tag.SIZE, len(tag))

    def test_tag_with_invalid_length_raises_valueerror(self):
        self.assertRaises(ValueError, Tag, '123456789')

    def test_tag_has_correct_data(self):
        tag = Tag(b'foo')
        self.assertEqual(b'foo\x00\x00\x00\x00\x00', tag.data)
        tag = Tag('abcdefgh')
        self.assertEqual('abcdefgh', tag.data)

    def test_tag_has_correct_string_representation(self):
        tag = Tag(b'foo')
        self.assertEqual('foo', str(tag))


class TestPrefix(TestCase):
    def test_init(self):
        Prefix(b'foo', 1)


class TestMessage(TestCase):
    def test_init(self):
        Message('')



try:
    import urwid
except ImportError:
    print("Could not import the Python module 'urwid'.")
    raise SystemExit


def handle_input(input):
    if input in UI.keys['escape']:
        raise urwid.ExitMainLoop()
    if input in UI.keys['left'] + UI.keys['right']:
        main_frame.set_focus('footer')
        next_blob()
    if input in UI.keys['down'] + UI.keys['up']:
        main_frame.set_focus('body')


class UI(object):
    """Represents the settings for the UI."""
    fg = 'light gray'
    bg = 'black'

    palette = [
        ('default', fg, bg),
        ('highlight', fg + ',standout', bg),
        ('header', 'white', 'dark cyan'),
        ('footer', 'light gray', 'dark blue'),
        ('body', 'dark cyan', '', 'standout'),
        ('focus', 'dark red', '', 'standout'),
        ('head', 'light red', 'black'),
        ('blob', 'yellow', 'dark cyan'),
        ('blob_scale', 'dark cyan', 'black'),
    ]

    keys = {
        'select': ('return', 'enter'),
        'inspect': ('x', 'X'),
        'escape': ('esc', 'q', 'Q'),
        'left': ('left', 'h'),
        'right': ('right', 'l'),
        'up': ('up', 'k'),
        'down': ('down', 'j'),
        'goto': ('g', 'G'),
        'help': ('?', ),
    }


class ItemWidget(urwid.WidgetWrap):
    def __init__(self, id, description):
        self.id = id
        self.content = 'item %s: %s...' % (str(id), description[:25])
        self.item = [
            (
                'fixed', 15,
                urwid.Padding(
                    urwid.AttrWrap(
                        urwid.Text('item %s' % str(id)), 'body', 'focus'
                    ),
                    left=2
                )
            ),
            urwid.AttrWrap(urwid.Text('%s' % description), 'body', 'focus'),
        ]
        w = urwid.Columns(self.item)
        self.__super.__init__(w)

    def selectable(self):
        return True

    def keypress(self, size, key):
        return key


class BlobSelector(urwid.WidgetWrap):
    def __init__(self, description):
        self.content = description
        self.item = [
            urwid.AttrWrap(urwid.Text('%s' % description), 'blob', 'focus'),
        ]
        w = urwid.Columns(self.item)
        self.__super.__init__(w)

    def selectable(self):
        return True

    def keypress(self, size, key):
        return key


class BlobWidget(urwid.Pile):
    def __init__(self):
        self.width = 20
        self.size = (0, )
        urwid.Pile.__init__(
            self, [
                urwid.Text('', wrap='clip'),
                urwid.Text('', wrap='clip'),
                urwid.Text('', wrap='clip')
            ]
        )

    def draw(self):
        self.widget_list[0].set_text(".OOOOOOOOOOOOOOOOOOOO")
        self.widget_list[1].set_text("|    '    |    '    |")
        self.widget_list[2].set_text("0         10        20")

    def render(self, size, focus):
        self.size = size
        self.draw()
        return urwid.Pile.render(self, size, focus)


def next_blob():
    pass


header = urwid.AttrWrap(urwid.Text("The header!", align='center'), 'header')
footer = urwid.AttrWrap(urwid.Text("The footer"), 'footer')

items = []
for i in range(100):
    items.append(ItemWidget(i, "Item {0}".format(i)))

browser_header = urwid.AttrMap(urwid.Text('selected:'), 'head')
browser_listbox = urwid.ListBox(urwid.SimpleListWalker(items))
browser_view = urwid.Frame(
    urwid.AttrWrap(browser_listbox, 'body'), header=browser_header
)

blobs = BlobWidget()
footer = urwid.Columns([urwid.Text('Info\nSecond kube'), blobs])

main_frame = urwid.AttrWrap(
    urwid.Frame(browser_view, focus_part='body'), 'default'
)
main_frame.set_header(header)
main_frame.set_footer(footer)

loop = urwid.MainLoop(main_frame, UI.palette, unhandled_input=handle_input)
loop.run()



# -*- coding: utf-8 -*-
"""

==================
Colormap pitfalls.
==================

Choosing the right colormap is important. A bad colormap like ``jet``
(standard in ROOT and matplotlib < 2.0) fools you into seeing structure
where there isn't any.
"""

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import km3pipe.style
km3pipe.style.use("km3pipe")

x = np.linspace(-np.pi, np.pi, 300)
xx, yy = np.meshgrid(x, x)
z = np.cos(xx) + np.cos(yy)

fig, axes = plt.subplots(2, 2)

for ax, colourmap in zip(axes.flatten(),
                         ['gray', 'jet', 'cubehelix', 'cividis']):
    ax.set_title(colourmap)
    im = ax.imshow(z, cmap=plt.get_cmap(colourmap))
    divider = make_axes_locatable(ax)
    cax = divider.append_axes('right', size='7%', pad=0.05)
    cbar = plt.colorbar(im, cax=cax)



# Filename: style.py
# pylint: disable=locally-disabled
"""
The KM3Pipe style definitions.

"""

from os.path import dirname, join, exists
from itertools import cycle

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

STYLE_DIR = join(dirname(dirname(__file__)), 'stylelib')


def get_style_path(style):
    return STYLE_DIR + '/' + style + '.mplstyle'


def use(style='km3pipe'):
    import matplotlib.pyplot as plt

    for s in (get_style_path('km3pipe-' + style), get_style_path(style),
              style):
        if exists(s):
            plt.style.use(s)
            return
    print("Could not find style: '{0}'".format(style))


class ColourCycler(object):
    """Basic colour cycler.

    Instantiate with `cc = ColourCycler()` and use it in plots
    like `plt.plot(xs, ys, c=next(cc))`.
    """
    def __init__(self, palette='km3pipe'):
        self.colours = {}
        self.refresh_styles()
        self.choose(palette)

    def choose(self, palette):
        """Pick a palette"""
        try:
            self._cycler = cycle(self.colours[palette])
        except KeyError:
            raise KeyError(
                "Chose one of the following colour palettes: {0}".format(
                    self.available
                )
            )

    def refresh_styles(self):
        """Load all available styles"""
        import matplotlib.pyplot as plt

        self.colours = {}
        for style in plt.style.available:
            try:
                style_colours = plt.style.library[style]['axes.prop_cycle']
                self.colours[style] = [c['color'] for c in list(style_colours)]
            except KeyError:
                continue

        self.colours['km3pipe'] = [
            "#ff7869", "#4babe1", "#96ad3e", "#e4823d", "#5d72b2", "#e2a3c2",
            "#fd9844", "#e480e7"
        ]

    @property
    def available(self):
        """Return a list of available styles"""
        return list(self.colours.keys())

    def __next__(self):
        """Return the next colour in current palette"""
        return next(self._cycler)

    def next(self):
        """Python 2 compatibility for iterators"""
        return self.__next__()



# Filename: test_mc.py
# pylint: disable=C0111,R0904,C0103
# vim:set ts=4 sts=4 sw=4 et:
"""
Tests for MC Modules.
"""

import numpy as np
from numpy.testing import (assert_array_equal, assert_allclose)
import pytest

from km3pipe import Table, Blob, Pipeline, Module
from km3pipe.testing import TestCase

from km3modules.mc import (
    convert_mc_times_to_jte_times, MCTimeCorrector, GlobalRandomState
)

__author__ = "Moritz Lotze, Michael Moser"
__copyright__ = "Copyright 2018, Tamas Gal and the KM3NeT collaboration."
__license__ = "MIT"
__maintainer__ = "Tamas Gal, Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class TestMCConvert(TestCase):
    def setUp(self):
        self.event_info = Table({
            'timestamp': 1,
            'nanoseconds': 700000000,
            'mc_time': 1.74999978e9,
        })

        self.mc_tracks = Table({
            'time': 1,
        })

        self.mc_hits = Table({
            'time': 30.79,
        })

        self.blob = Blob({
            'event_info': self.event_info,
            'mc_hits': self.mc_hits,
            'mc_tracks': self.mc_tracks,
        })

    def test_convert_mc_times_to_jte_times(self):
        times_mc_tracks = convert_mc_times_to_jte_times(
            self.mc_tracks.time,
            self.event_info.timestamp * 1e9 + self.event_info.nanoseconds,
            self.event_info.mc_time
        )
        times_mc_hits = convert_mc_times_to_jte_times(
            self.mc_hits.time,
            self.event_info.timestamp * 1e9 + self.event_info.nanoseconds,
            self.event_info.mc_time
        )

        assert times_mc_tracks is not None
        assert times_mc_hits is not None
        print(times_mc_tracks, times_mc_hits)
        assert np.allclose(times_mc_tracks, 49999781)
        assert np.allclose(times_mc_hits, 49999810.79)

    def test_process(self):
        corr = MCTimeCorrector(
            mc_hits_key='mc_hits',
            mc_tracks_key='mc_tracks',
            event_info_key='event_info'
        )
        newblob = corr.process(self.blob)
        assert newblob['mc_hits'] is not None
        assert newblob['mc_tracks'] is not None
        assert np.allclose(newblob['mc_hits'].time, 49999810.79)
        assert np.allclose(newblob['mc_tracks'].time, 49999781)


class TestGlobalRandomState(TestCase):
    def test_default_random_state(self):
        assertAlmostEqual = self.assertAlmostEqual

        class Observer(Module):
            def configure(self):
                self.i = 0
                self.x = [0.3745401188, 0.950714306, 0.7319939418]

            def process(self, blob):
                assertAlmostEqual(self.x[self.i], np.random.rand())
                self.i += 1
                return blob

        pipe = Pipeline()
        pipe.attach(GlobalRandomState)
        pipe.attach(Observer)
        pipe.drain(3)

    def test_custom_random_state(self):
        assertAlmostEqual = self.assertAlmostEqual

        class Observer(Module):
            def configure(self):
                self.i = 0
                self.x = [0.221993171, 0.870732306, 0.206719155]

            def process(self, blob):
                assertAlmostEqual(self.x[self.i], np.random.rand())
                self.i += 1
                return blob

        pipe = Pipeline()
        pipe.attach(GlobalRandomState, seed=5)
        pipe.attach(Observer)
        pipe.drain(3)

    def test_without_pipeline_and_default_state(self):
        GlobalRandomState()
        numbers = np.arange(1, 50)
        np.random.shuffle(numbers)
        lotto_numbers = sorted(numbers[:6])
        self.assertListEqual([14, 18, 28, 45, 46, 48], lotto_numbers)

    def test_without_pipeline_with_custom_seed(self):
        GlobalRandomState(seed=23)
        numbers = np.arange(1, 50)
        np.random.shuffle(numbers)
        lotto_numbers = sorted(numbers[:6])
        self.assertListEqual([14, 15, 18, 19, 33, 44], lotto_numbers)



from km3pipe import Pipeline, Module
from km3pipe.io import CHPump


class CHPrinter(Module):
    def process(self, blob):
        print("New blob:")
        print(blob['CHPrefix'])
        print(blob['CHData'])
        return blob


pipe = Pipeline()
pipe.attach(
    CHPump,
    host='127.0.0.1',
    port=5553,
    tags="foo, narf",
    timeout=1000,
    max_queue=42
)
pipe.attach(CHPrinter)
pipe.drain()



"""
Print the ROOT file structure.

Usage:
    rtree [options] FILE
    rtree (-h | --help)
    rtree --version

Options:
    FILE       Input file.
    -h --help  Show this screen.

"""
from km3pipe.io.root import open_rfile

__author__ = "Moritz Lotze"
__copyright__ = "Copyright 2016, Moritz Lotze and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Moritz Lotze"
__email__ = "mlotze@km3net.de"
__status__ = "Development"


def rtree(rfile):
    rfile = open_rfile(rfile)
    print(rfile.ls())
    rfile.close()


def main():
    from docopt import docopt
    arguments = docopt(__doc__)

    rtree(arguments['FILE'])



import numpy as np
import km3pipe as kp


class APump(kp.Pump):
    def configure(self):
        self.index = 0

    def process(self, blob):
        data = {'a': self.index * np.arange(5), 'b': np.arange(5)**self.index}
        data2 = {
            'c': self.index * np.arange(10, dtype='f4') + 0.1,
            'd': np.arange(10, dtype='f4')**self.index + 0.2
        }
        print(data2)
        blob['Tablelike'] = kp.Table(data, h5loc='/tablelike', name='2D Table')
        print(blob['Tablelike'])
        blob['Columnwise'] = kp.Table(
            data2,
            h5loc='/columnwise',
            split_h5=True,
            name='Column-wise Split'
        )
        self.index += 1
        return blob


pipe = kp.Pipeline()
pipe.attach(APump)
pipe.attach(kp.io.HDF5Sink, filename='km3hdf5_example.h5')
pipe.drain(13)



# Filename: constants.py
# pylint: disable=C0103
# pragma: no cover
"""
The constants used in KM3Pipe.

"""
import math

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

# Detector related parameters
arca_frame_duration = 0.1    # s
orca_frame_duration = 0.1    # s

c = 2.99792458e8    # m/s

n_water_antares_phase = 1.3499
n_water_antares_group = 1.3797
n_water_km3net_group = 1.3787
n_water_antares = n_water_antares_group
theta_cherenkov_water_antares = math.acos(1 / n_water_antares_phase)
theta_cherenkov_water_km3net = math.acos(1 / n_water_km3net_group)
c_water_antares = c / n_water_antares_group
c_water_km3net = c / n_water_km3net_group

# Math
pi = math.pi
e = math.e

# Default values for time residuals
dt_window_l = -15    # ns
dt_window_h = +25    # ns



#!/usr/bin/env python
# vim: ts=4 sw=4 et
"""
===========
HRV Monitor
===========

The following script checks the high-rate-veto for each PMT.

"""

# Author: Tamas Gal <tgal@km3net.de>
# License: MIT
from datetime import datetime
import io
from collections import defaultdict
import threading
import time
import km3pipe as kp
from km3pipe.io.daq import TMCHData
import numpy as np
import matplotlib
matplotlib.use('Agg')    # noqa
import matplotlib.pyplot as plt
import km3pipe.style as kpst
kpst.use("km3pipe")

__author__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
VERSION = "1.0"

log = kp.logger.get_logger("HRV")


class PMTRates(kp.Module):
    def configure(self):
        self.detector = self.require("detector")
        self.du = self.require("du")
        self.interval = self.get("interval") or 10
        self.plot_path = self.get("plot_path") or "km3web/plots/hrv.png"
        self.max_x = 800
        self.index = 0
        self.hrv = defaultdict(list)
        self.hrv_matrix = np.full((18 * 31, self.max_x), np.nan)
        self.lock = threading.Lock()
        self.thread = threading.Thread(target=self.run, args=())
        self.thread.daemon = True
        self.thread.start()

    def run(self):
        interval = self.interval
        while True:
            time.sleep(interval)
            now = datetime.now()
            self.add_column()
            self.update_plot()
            with self.lock:
                self.hrv = defaultdict(list)
            delta_t = (datetime.now() - now).total_seconds()
            remaining_t = self.interval - delta_t
            print(
                "Delta t: {} -> waiting for {}s".format(
                    delta_t, self.interval - delta_t
                )
            )
            if (remaining_t < 0):
                log.error(
                    "Can't keep up with plot production. "
                    "Increase the interval!"
                )
                interval = 1
            else:
                interval = remaining_t

    def add_column(self):
        m = np.roll(self.hrv_matrix, -1, 1)
        y_range = 18 * 31
        mean_hrv = np.full(y_range, np.nan)
        for i in range(y_range):
            if i not in self.hrv:
                continue
            mean_hrv[i] = np.mean(self.hrv[i])

        m[:, self.max_x - 1] = mean_hrv
        self.hrv_matrix = m
        print(self.hrv_matrix)

    def update_plot(self):
        print("Updating plot at {}".format(self.plot_path))
        now = time.time()
        max_x = self.max_x
        interval = self.interval

        def xlabel_func(timestamp):
            return datetime.utcfromtimestamp(timestamp).strftime("%H:%M")

        m = self.hrv_matrix
        fig, ax = plt.subplots(figsize=(10, 6))
        ax.imshow(m, origin='lower')
        ax.set_title(
            "HRV Ratios for DU-{}\n{}".format(self.du, datetime.utcnow())
        )
        ax.set_xlabel("UTC time [{}s/px]".format(interval))
        plt.yticks([i * 31 for i in range(18)],
                   ["Floor {}".format(f) for f in range(1, 19)])
        xtics_int = range(0, max_x, int(max_x / 10))
        plt.xticks([i for i in xtics_int], [
            xlabel_func(now - (max_x - i) * interval) for i in xtics_int
        ])
        fig.tight_layout()
        plt.savefig(self.plot_path)
        plt.close('all')

    def process(self, blob):
        tmch_data = TMCHData(io.BytesIO(blob['CHData']))
        dom_id = tmch_data.dom_id

        if dom_id not in self.detector.doms:
            return blob

        du, floor, _ = self.detector.doms[dom_id]

        if du != self.du:
            return blob

        hrv_flags = reversed("{0:b}".format(tmch_data.hrvbmp).zfill(32))

        y_base = (floor - 1) * 31

        for channel_id, hrv_flag in enumerate(hrv_flags):
            idx = y_base + channel_id
            with self.lock:
                self.hrv[idx].append(int(hrv_flag))

        return blob


def main():
    detector = kp.hardware.Detector(det_id=29)
    pipe = kp.Pipeline(timeit=True)
    pipe.attach(
        kp.io.CHPump,
        host='192.168.0.110',
        port=5553,
        tags='IO_MONIT',
        timeout=60 * 60 * 24 * 7,
        max_queue=1000
    )
    pipe.attach(PMTRates, detector=detector, du=2, interval=10)
    pipe.drain()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
Converts hits in a Jpp-ROOT file to HDF5.

"""
from km3pipe import Pipeline
from km3pipe.io.aanet import AanetPump
from km3pipe.io import HDF5SinkLegacy

import sys

if len(sys.argv) < 3:
    sys.exit('Usage: {0} FILENAME.root OUTPUTFILENAME.h5'.format(sys.argv[0]))

FILEPATH = sys.argv[1]
OUTPUTFILEPATH = sys.argv[2]

pipe = Pipeline()
pipe.attach(AanetPump, filename=FILEPATH)
pipe.attach(HDF5SinkLegacy, filename=OUTPUTFILEPATH)
pipe.drain()



# Filename: sys.py
# pylint: disable=C0103
"""
Some unsorted, frequently used logic.

"""
try:
    import resource    # linux/macos
except ImportError:
    import psutil    # windows

import sys
from contextlib import contextmanager

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal and Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"


@contextmanager
def ignored(*exceptions):
    """Ignore-context for a given list of exceptions.

    Example:
        with ignored(AttributeError):
            foo.a = 1

    """
    try:
        yield
    except exceptions:
        pass


def peak_memory_usage():
    """Return peak memory usage in MB"""
    if sys.platform.startswith('win'):
        p = psutil.Process()
        return p.memory_info().peak_wset / 1024 / 1024

    mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
    factor_mb = 1 / 1024
    if sys.platform == 'darwin':
        factor_mb = 1 / (1024 * 1024)
    return mem * factor_mb



# Filename: test_plot.py
# pylint: disable=locally-disabled,C0111,R0904,C0103

import numpy as np

from km3pipe.testing import TestCase, patch
from km3pipe.plot import meshgrid, automeshgrid, diag

__author__ = "Moritz Lotze"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Moritz Lotze"
__email__ = "mlotze@km3net.de"
__status__ = "Development"


class TestMeshStuff(TestCase):
    def test_meshgrid(self):
        xx, yy = meshgrid(-1, 1, 0.8)
        assert np.allclose([[-1.0, -0.2, 0.6], [-1.0, -0.2, 0.6],
                            [-1.0, -0.2, 0.6]], xx)
        assert np.allclose([[-1.0, -1.0, -1.0], [-0.2, -0.2, -0.2],
                            [0.6, 0.6, 0.6]], yy)

    def test_meshgrid_with_y_specs(self):
        xx, yy = meshgrid(-1, 1, 0.8, -10, 10, 8)
        assert np.allclose([[-1.0, -0.2, 0.6], [-1.0, -0.2, 0.6],
                            [-1.0, -0.2, 0.6]], xx)
        assert np.allclose([[-10, -10, -10], [-2, -2, -2], [6, 6, 6]], yy)



#!/usr/bin/env python
# Filename: setup.py
"""
KM3Pipe setup script.

"""

from setuptools import setup
import sys

import builtins
# so we can detect in __init__.py that it's called from setup.py
builtins.__KM3PIPE_SETUP__ = True

with open('requirements.txt') as fobj:
    requirements = [l.strip() for l in fobj.readlines()]

with open('requirements-dev.txt') as fobj:
    requirements_dev = [l.strip() for l in fobj.readlines()]

try:
    with open("README.rst") as fh:
        long_description = fh.read()
except UnicodeDecodeError:
    long_description = "KM3Pipe"

setup(
    name='km3pipe',
    url='http://git.km3net.de/km3py/km3pipe',
    description='An analysis framework for KM3NeT',
    long_description=long_description,
    author='Tamas Gal and Moritz Lotze',
    author_email='tgal@km3net.de',
    packages=[
        'km3pipe', 'km3pipe.io', 'km3pipe.utils', 'km3modules', 'pipeinspector'
    ],
    include_package_data=True,
    platforms='any',
    setup_requires=['numpy>=1.12', 'setuptools_scm'],
    install_requires=requirements,
    extras_require = {"dev": requirements_dev},
    use_scm_version=True,
    python_requires='>=3.5',
    entry_points={
        'console_scripts': [
            'km3pipe=km3pipe.cmd:main',
            'pipeinspector=pipeinspector.app:main',
            'rtree=km3pipe.utils.rtree:main',
            'h5info=km3pipe.utils.h5info:main',
            'h5tree=km3pipe.utils.h5tree:main',
            'h5header=km3pipe.utils.h5header:main',
            'ptconcat=km3pipe.utils.ptconcat:main',
            'meantots=km3pipe.utils.meantots:main',
            'pushover=km3pipe.utils.pushover:main',
            'ztplot=km3pipe.utils.ztplot:main',
            'k40calib=km3pipe.utils.k40calib:main',
            'totmonitor=km3pipe.utils.totmonitor:main',
            'streamds=km3pipe.utils.streamds:main',
            'triggermap=km3pipe.utils.triggermap:main',
            'nb2sphx=km3pipe.utils.nb2sphx:main',
            'triggersetup=km3pipe.utils.triggersetup:main',
            'ligiermirror=km3pipe.utils.ligiermirror:main',
            'runtable=km3pipe.utils.runtable:main',
            'runinfo=km3pipe.utils.runinfo:main',
            'qrunprocessor=km3pipe.utils.qrunprocessor:main',
            'wtd=km3pipe.utils.wtd:main',
            'qrunqaqc=km3pipe.utils.qrunqaqc:main',
            'daqsample=km3pipe.utils.daqsample:main',
        ],
    },
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python',
    ],
)

__author__ = 'Tamas Gal and Moritz Lotze'



# Filename: controlhost.py
"""
A set of classes and tools wich uses the ControlHost protocol.

"""
import socket
import struct
import time

from .logger import get_logger

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

log = get_logger(__name__)

BUFFER_SIZE = 1024


class Client(object):
    """The ControlHost client"""
    def __init__(self, host, port=5553):
        self.host = host
        self.port = port
        self.socket = None
        self.tags = []
        self.valid_tags = []

    def subscribe(self, tag, mode='wait'):
        if mode not in ['wait', 'all']:
            raise ValueError("Possible subscription modes are 'wait' or 'all'")
        log.info("Subscribing to {} in mode {}".format(tag, mode))
        full_tag = self._full_tag(tag, mode)
        if full_tag not in self.tags:
            self.tags.append(full_tag)
        for t in tag.split():
            if t not in self.valid_tags:
                self.valid_tags.append(t)
        self._update_subscriptions()

    def unsubscribe(self, tag, mode='wait'):
        try:
            self.tags.remove(self._full_tag(tag, mode))
            self.valid_tags.remove(tag)
        except ValueError:
            pass
        else:
            self._update_subscriptions()

    def _full_tag(self, tag, mode):
        mode_flag = ' {} '.format(mode[0])
        full_tag = mode_flag + tag
        return full_tag

    def _update_subscriptions(self):
        log.debug("Subscribing to tags: {0}".format(self.tags))
        if not self.socket:
            self._connect()
        tags = ''.join(self.tags).encode("ascii")
        message = Message(b'_Subscri', tags)
        self.socket.send(message.data)
        message = Message(b'_Always')
        self.socket.send(message.data)

    def put_message(self, tag, data):
        """Send data to the ligier with a given tag"""
        if not self.socket:
            self._connect()
        msg = Message(tag, data)
        self.socket.send(msg.data)

    def get_message(self):
        while True:
            log.info("     Waiting for control host Prefix")
            try:
                data = self.socket.recv(Prefix.SIZE)
                timestamp = time.time()
                log.info("    raw prefix data received: '{0}'".format(data))
                if data == b'':
                    raise EOFError
                prefix = Prefix(data=data, timestamp=timestamp)
            except (UnicodeDecodeError, OSError, struct.error):
                log.error("Failed to construct Prefix, reconnecting.")
                self._reconnect()
                continue

            try:
                prefix_tag = str(prefix.tag)
            except UnicodeDecodeError:
                log.error("The tag could not be decoded. Reconnecting.")
                self._reconnect()
                continue

            if prefix_tag not in self.valid_tags:
                log.error(
                    "Invalid tag '{0}' received, ignoring the message \n"
                    "and reconnecting.\n"
                    "  -> valid tags are: {1}".format(
                        prefix_tag, self.valid_tags
                    )
                )
                self._reconnect()
                continue
            else:
                break

        message = b''
        log.info("       got a Prefix with {0} bytes.".format(prefix.length))
        while len(message) < prefix.length:
            log.info("          message length: {0}".format(len(message)))
            log.info("            (getting next part)")
            buffer_size = min((BUFFER_SIZE, (prefix.length - len(message))))
            try:
                message += self.socket.recv(buffer_size)
            except OSError:
                log.error("Failed to construct message.")
                raise BufferError
        log.info(
            "     ------ returning message with {0} bytes".format(
                len(message)
            )
        )
        return prefix, message

    def _connect(self):
        """Connect to JLigier"""
        log.debug("Connecting to JLigier")
        self.socket = socket.socket()
        self.socket.connect((self.host, self.port))

    def _disconnect(self):
        """Close the socket"""
        log.debug("Disconnecting from JLigier")
        if self.socket:
            self.socket.close()

    def _reconnect(self):
        """Reconnect to JLigier and subscribe to the tags."""
        log.debug("Reconnecting to JLigier...")
        self._disconnect()
        self._connect()
        self._update_subscriptions()

    def __enter__(self):
        self._connect()
        return self

    def __exit__(self, exit_type, value, traceback):
        self._disconnect()


class Message(object):
    """The representation of a ControlHost message."""
    def __init__(self, tag, message=b''):
        try:
            message = message.encode()
        except AttributeError:
            pass
        try:
            tag = tag.encode()
        except AttributeError:
            pass
        self.prefix = Prefix(tag, len(message))
        self.message = message

    @property
    def data(self):
        return self.prefix.data + self.message


class Tag(object):
    """Represents the tag in a ControlHost Prefix."""
    SIZE = 8

    def __init__(self, data=None):
        self._data = b''
        self.data = data

    @property
    def data(self):
        """The byte data"""
        return self._data

    @data.setter
    def data(self, value):
        """Set the byte data and fill up the bytes to fit the size."""
        if not value:
            value = b''
        if len(value) > self.SIZE:
            raise ValueError("The maximum tag size is {0}".format(self.SIZE))
        self._data = value
        while len(self._data) < self.SIZE:
            self._data += b'\x00'

    def __str__(self):
        return self.data.decode(encoding='UTF-8').strip('\x00')

    def __len__(self):
        return len(self._data)


class Prefix(object):
    """The prefix of a ControlHost message."""
    SIZE = 16

    def __init__(self, tag=None, length=None, data=None, timestamp=None):
        if data:
            self.data = data
        else:
            self.tag = Tag(tag)
            self.length = length
        if timestamp is None:
            self.timestamp = time.time()
        else:
            self.timestamp = timestamp

    @property
    def data(self):
        return self.tag.data + struct.pack('>i', self.length) + b'\x00' * 4

    @data.setter
    def data(self, value):
        self.tag = Tag(data=value[:Tag.SIZE])
        self.length = struct.unpack('>i', value[Tag.SIZE:Tag.SIZE + 4])[0]

    def __str__(self):
        return (
            "ControlHost Prefix with tag '{0}' ({1} bytes of data)".format(
                self.tag, self.length
            )
        )



# Filename: __init__.py
"""
The extemporary KM3NeT analysis framework.

"""
from pkg_resources import get_distribution, DistributionNotFound

try:
    version = get_distribution(__name__).version
except DistributionNotFound:
    version = "unknown version"

try:
    __KM3PIPE_SETUP__
except NameError:
    __KM3PIPE_SETUP__ = False

if not __KM3PIPE_SETUP__:
    from . import logger    # noqa
    from .core import (Pipeline, Module, Pump, Blob)    # noqa
    from . import core    # noqa
    from .dataclasses import Table, NDArray    # noqa
    from . import dataclasses    # noqa
    from . import calib    # noqa
    from . import cmd    # noqa
    from . import config    # noqa
    from . import constants    # noqa
    from . import controlhost    # noqa
    from . import db    # noqa
    from . import hardware    # noqa
    from . import io    # noqa
    from . import math    # noqa
    from . import mc    # noqa
    from . import shell    # noqa
    from . import style    # noqa
    from . import sys    # noqa
    # from . import testing     # noqa
    from . import time    # noqa
    from . import tools    # noqa

__author__ = "Tamas Gal and Moritz Lotze"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = ["Thomas Heid"]
__license__ = "MIT"
__version__ = version
__maintainer__ = "Tamas Gal and Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"



# Filename: pushover.py
"""
Send a push message to a device using Pushover.net API.

Usage:
    pushover MESSAGE...
    pushover (-h | --help)
    pushover --version

Options:
    MESSAGE     The message to send.
    -h --help   Show this screen.

"""

from km3pipe.config import Config
from km3pipe import version

import http.client as httplib
from urllib.parse import urlencode

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2017, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


def main():
    from docopt import docopt
    args = docopt(__doc__, version=version)

    config = Config()

    token = config.get("Pushover", "token")
    user_key = config.get("Pushover", "user_key")
    if token is None or user_key is None:
        print(
            "Please define your 'token' and 'user_key' in the "
            "'Pushover' section of your ~/.km3net configuration."
        )
        exit(1)

    conn = httplib.HTTPSConnection("api.pushover.net:443")
    conn.request(
        "POST", "/1/messages.json",
        urlencode({
            "token": token,
            "user": user_key,
            "message": ' '.join(args["MESSAGE"]),
        }), {"Content-type": "application/x-www-form-urlencoded"}
    )

    conn.getresponse()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
I3Pump for ANTARES DATA.
"""
import numpy as np
import km3pipe as kp
from km3pipe.logger import get_logger

from icecube import icetray, dataclasses, dataio, antares_common
from icecube.dataclasses import *

__author__ = "Nicole Geisselbrecht"
__copyright__ = "Copyright 2016, Nicole Geisselbrecht and the KM3NeT collaboration."
__credits__ = "Tamas Gal"
__license__ = "MIT"
__maintainer__ = "Nicole Geisselbrecht"
__email__ = "ngeisselbrecht@km3net.de"
__status__ = "Development"

log = get_logger(__name__)    # pylint: disable=C0103


class I3Pump(kp.Module):
    """
        Gets information of an ANTARES I3 file and converts it to HDF5.

        Attributes
        ----------
        filename : str
            Path to the input I3 file.

    """
    def configure(self):
        self.filename = self.require("filename")
        self.fobj = dataio.I3File(self.filename)
        self.blobs = self.blob_generator()
        self.omGeoMap = None
        self.cog = None

        self._hit_buffer = np.empty((10000, 14))
        self._mchit_buffer = np.empty((10000, 12))

    def process(self, blob):
        """

        Parameters
        ----------
        blob : dict
            Empty dictionary.


        Returns
        -------
        blob : dict
            Dictionary which contains information of the I3 file.

        """

        frame = next(self.blobs)

        # EVENT HEADER INFO
        if 'I3EventHeader' in frame:
            header = frame['I3EventHeader']
            blob['EventInfo'] = self._read_event_info(frame, header)

        # MC INFO
        if 'AntMCTree' and 'I3MCWeightDict' in frame:

            weights = kp.Table({
                'w2': self.get_antares_w2(frame),
                'w3': self.get_antares_w3(frame),
                'nb_gen_events': self.get_nb_generated_events(frame)
            },
                               h5loc="/mc_info")

            blob['McInfo'] = weights

            nu = self.get_mc_primary(frame)
            muon = self.get_mc_highest_energy_muon(frame)

            if nu is not None:
                blob['NuInfo'] = self._read_particle_info(nu, "/nu")
            if muon is not None:
                blob['MuonInfo'] = self._read_particle_info(muon, "/muon")

        # MC HITS
        if 'MCHits' in frame:
            mchits = frame['MCHits']
            blob['McHits'] = self._read_mchits(frame, mchits)

        # AAFIT
        if 'AafitFinalFit' in frame:
            aafit = frame['AafitFinalFit']
            if aafit.fit_status == dataclasses.I3Particle.FitStatus.OK:
                blob['AafitInfo'] = self._read_aafit(frame, aafit)

        # BBFIT
        if 'BBFitTrack' in frame:
            bbfit_track = frame['BBFitTrack']
            if bbfit_track.fit_status == dataclasses.I3Particle.FitStatus.OK:
                bbfit_track_chi2 = -1.
                if 'BBFitTChi2' in frame:
                    bbfit_track_chi2 = frame['BBFitTChi2']

                blob['BbfitTrackInfo'] = self._read_bbfit(
                    bbfit_track, bbfit_track_chi2, "/reco/bbfit_track"
                )

        if 'BBFitBright' in frame:
            bbfit_bright = frame['BBFitBright']
            if bbfit_bright.fit_status == dataclasses.I3Particle.FitStatus.OK:
                bbfit_bright_chi2 = -1.
                if 'BBFitBChi2' in frame:
                    bbfit_bright_chi2 = frame['BBFitBChi2']

                blob['BbfitBrightInfo'] = self._read_bbfit(
                    bbfit_bright, bbfit_bright_chi2, "/reco/bbfit_bright"
                )

        # GRIDFIT
        if 'GridFit_FinalFitResult' in frame:
            gridfit = frame['GridFit_FinalFitResult']
            if gridfit.fit_status == dataclasses.I3Particle.FitStatus.OK:
                blob['GridFitInfo'] = self._read_gridfit(
                    gridfit, "/reco/gridfit"
                )

        # HIT INFO
        if 'CalibratedPulses' in frame:
            pulses = frame['CalibratedPulses']
            blob['HitsInfo'] = self._read_hits(frame, pulses)

        return blob

    @staticmethod
    def _read_event_info(frame, header):
        """
        Reads event information of the I3 file and returns it as dict.
        """
        triggercounter = -1
        if 'AntaresTriggerCounter' in frame:
            trig = frame['AntaresTriggerCounter']
            triggercounter = trig.value

        triggermask = 0
        if 'AntTriggerHierarchy' in frame:
            trigger_types = {
                'ANT_TRIGGER_3D': 1,
                'ANT_TRIGGER_1D': 2,
                'ANT_TRIGGER_1D_WITH_PREFIT': 3,
                'ANT_TRIGGER_1D_MIXED': 4,
                'ANT_TRIGGER_1D_MIXED_WITH_PREFIT': 5,
                'ANT_TRIGGER_3D_SCAN': 6,
                'ANT_TRIGGER_3D_RECURSIVE': 7,
                'ANT_TRIGGER_OB': 8,
                'ANT_TRIGGER_MINIMUM_BIAS': 9,
                'ANT_TRIGGER_1S': 10,
                'ANT_TRIGGER_3S': 11,
                'ANT_TRIGGER_3S_SCAN': 12,
                'ANT_TRIGGER_TIMESTAMP_0': 13,
                'ANT_TRIGGER_T3': 14,
                'ANT_TRIGGER_T2': 15,
                'ANT_TRIGGER_TQ': 16,
                'ANT_TRIGGER_GC': 30,
                'ANT_MERGED': 32
            }
            tree = frame['AntTriggerHierarchy']
            for item in tree:
                trig_type = str(item.Key.Type)
                if trig_type in trigger_types:
                    triggermask = triggermask + 2**(
                        trigger_types[trig_type] - 1
                    )

        event_info = kp.Table({
            'runID': header.RunID,
            'eventID': header.EventID,
            'triggercounter': triggercounter,
            'startTimeUnix': str(header.StartTime.GetUnixTime()),
            'startTime': str(header.StartTime),
            'triggermask': triggermask
        },
                              h5loc="/event_info")
        return event_info

    @staticmethod
    def _read_particle_info(particle, h5loc):
        """
        Reads particle information of the I3 file and returns it as dict.

        """
        particle_mc = kp.Table({
            'dir_x': particle.GetDir().GetX(),
            'dir_y': particle.GetDir().GetY(),
            'dir_z': particle.GetDir().GetZ(),
            'pos_x': particle.GetPos().X,
            'pos_y': particle.GetPos().Y,
            'pos_z': particle.GetPos().Z,
            'energy': particle.GetEnergy(),
            'pdgencoding': particle.GetPdgEncoding()
        },
                               h5loc=h5loc)
        return particle_mc

    @staticmethod
    def _read_aafit(frame, aafit):
        """
        Reads aafit information of the I3 file and returns it as dict.
        """
        lambda_value = -1
        beta_value = -1
        if 'AafitLambdaFinalFit' in frame:
            lambda_frame = frame['AafitLambdaFinalFit']
            lambda_value = lambda_frame.value

        if 'AafitErrorEstimateFinalFit' in frame:
            beta_frame = frame['AafitErrorEstimateFinalFit']
            beta_value = beta_frame.value

        aafit_pos = aafit.GetPos()
        aafit_dir = aafit.GetDir()

        aafit_info = kp.Table({
            'dir_x': aafit_dir.GetX(),
            'dir_y': aafit_dir.GetY(),
            'dir_z': aafit_dir.GetZ(),
            'pos_x': aafit_pos.X,
            'pos_y': aafit_pos.Y,
            'pos_z': aafit_pos.Z,
            'lambda': lambda_value,
            'beta': beta_value
        },
                              h5loc="/reco/aafit")
        return aafit_info

    @staticmethod
    def _read_bbfit(bbfit, bbfit_chi2, h5loc):
        """
        Reads bbfit information of the I3 file and returns it as dict.
        """
        bbfit_pos = bbfit.GetPos()
        bbfit_dir = bbfit.GetDir()

        bbfit_info = kp.Table({
            'dir_x': bbfit_dir.GetX(),
            'dir_y': bbfit_dir.GetY(),
            'dir_z': bbfit_dir.GetZ(),
            'pos_x': bbfit_pos.X,
            'pos_y': bbfit_pos.Y,
            'pos_z': bbfit_pos.Z,
            'chi2': bbfit_chi2.value
        },
                              h5loc=h5loc)
        return bbfit_info

    @staticmethod
    def _read_gridfit(gridfit, h5loc):
        """
        Reads gridfit information of the I3 file and returns it as dict.
        """
        gridfit_pos = gridfit.GetPos()
        gridfit_dir = gridfit.GetDir()
        gridfit_info = kp.Table({
            'dir_x': gridfit_dir.GetX(),
            'dir_y': gridfit_dir.GetY(),
            'dir_z': gridfit_dir.GetZ(),
            'pos_x': gridfit_pos.X,
            'pos_y': gridfit_pos.Y,
            'pos_z': gridfit_pos.Z,
            'lik': gridfit.GetLik()
        },
                                h5loc=h5loc)
        return gridfit_info

    def _read_hits(self, frame, pulses):
        """Creates a hits Table from calibrated pulses"""
        idx = 0
        if pulses is not None:
            for omkey, hitseries in pulses:
                omgeo = self.omGeoMap[omkey]
                rate = self.get_rate(frame, omkey)
                for hit in hitseries:

                    if idx == len(self._hit_buffer):
                        self.increase_hits_buffer_size()
                    floor = 1 + (omkey.om - 1) / 3
                    pmt = (omkey.om + 2) % 3

                    self._hit_buffer[idx][0] = hit.ID
                    self._hit_buffer[idx][1] = omkey.string
                    self._hit_buffer[idx][2] = floor
                    self._hit_buffer[idx][3] = pmt
                    self._hit_buffer[idx][4] = omgeo.position.X - self.cog.X
                    self._hit_buffer[idx][5] = omgeo.position.Y - self.cog.Y
                    self._hit_buffer[idx][6] = omgeo.position.Z - self.cog.Z
                    self._hit_buffer[idx][7] = omgeo.orientation.GetDirX()
                    self._hit_buffer[idx][8] = omgeo.orientation.GetDirY()
                    self._hit_buffer[idx][9] = omgeo.orientation.GetDirZ()
                    self._hit_buffer[idx][10] = hit.Time
                    self._hit_buffer[idx][11] = hit.Charge
                    self._hit_buffer[idx][12] = rate
                    self._hit_buffer[idx][13] = hit.IsTriggered()
                    idx += 1

            buf = self._hit_buffer[:idx]

            hits = kp.Table({
                'id': buf[:, 0].astype(int),
                'line': buf[:, 1].astype(np.uint8),
                'floor': buf[:, 2].astype(np.uint8),
                'pmt': buf[:, 3].astype(np.uint32),
                'x': buf[:, 4],
                'y': buf[:, 5],
                'z': buf[:, 6],
                'dx': buf[:, 7],
                'dy': buf[:, 8],
                'dz': buf[:, 9],
                'time': buf[:, 10].astype(np.float32),
                'charge': buf[:, 11],
                'rate': buf[:, 12],
                'triggered': buf[:, 13]
            },
                            h5loc="/hits",
                            split_h5=True)
            return hits

    def increase_hits_buffer_size(self):
        """
        Increases the size of the hit buffer.
        """
        current_buffer_size = len(self._hit_buffer)
        new_buffer_size = int(current_buffer_size * 1.3)
        new_buffer = np.empty((new_buffer_size, 14))
        new_buffer[:current_buffer_size] = self._hit_buffer
        self._hit_buffer = new_buffer

    def _read_mchits(self, frame, mchits):
        """Creates a mchits Table from MCHits"""
        idx = 0
        if mchits is not None:
            for omkey, hitseries in mchits:
                om_geo = self.omGeoMap[omkey]
                rate = self.get_rate(frame, omkey)
                for hit in hitseries:
                    if idx == len(self._mchit_buffer):
                        self.increase_mchits_buffer_size()
                    floor = 1 + (omkey.om - 1) / 3
                    pmt = (omkey.om + 2) % 3
                    self._mchit_buffer[idx][0] = hit.HitID
                    self._mchit_buffer[idx][1] = omkey.string
                    self._mchit_buffer[idx][2] = floor
                    self._mchit_buffer[idx][3] = pmt
                    self._mchit_buffer[idx][4] = om_geo.position.X - self.cog.X
                    self._mchit_buffer[idx][5] = om_geo.position.Y - self.cog.Y
                    self._mchit_buffer[idx][6] = om_geo.position.Z - self.cog.Z
                    self._mchit_buffer[idx][7] = om_geo.orientation.GetDirX()
                    self._mchit_buffer[idx][8] = om_geo.orientation.GetDirY()
                    self._mchit_buffer[idx][9] = om_geo.orientation.GetDirZ()
                    self._mchit_buffer[idx][10] = hit.Time
                    self._mchit_buffer[idx][11] = rate
                    idx += 1

            mcbuf = self._mchit_buffer[:idx]

            mchits = kp.Table({
                'id': mcbuf[:, 0].astype(int),
                'line': mcbuf[:, 1].astype(np.uint8),
                'floor': mcbuf[:, 2].astype(np.uint8),
                'pmt': mcbuf[:, 3].astype(np.uint32),
                'x': mcbuf[:, 4],
                'y': mcbuf[:, 5],
                'z': mcbuf[:, 6],
                'dx': mcbuf[:, 7],
                'dy': mcbuf[:, 8],
                'dz': mcbuf[:, 9],
                'time': mcbuf[:, 10].astype(np.float32),
                'rate': mcbuf[:, 11]
            },
                              h5loc="/mchits",
                              split_h5=True)
            return mchits

    def increase_mchits_buffer_size(self):
        """
        Increases the size of the mchit buffer.
        """
        current_mcbuffer_size = len(self._mchit_buffer)
        new_mcbuffer_size = int(current_mcbuffer_size * 1.3)
        new_mcbuffer = np.empty((new_mcbuffer_size, 12))
        new_mcbuffer[:current_mcbuffer_size] = self._mchit_buffer
        self._mchit_buffer = new_mcbuffer

    @staticmethod
    def get_mc_highest_energy_muon(frame):
        mctree = frame.Get("AntMCTree")
        mctracksinice = mctree.GetInIce()
        mctrack = None
        highest_energy = -1.

        for track in mctracksinice:
            if (track.GetType() == dataclasses.I3Particle.ParticleType.MuPlus) or \
                    (track.GetType() == dataclasses.I3Particle.ParticleType.MuMinus):
                if track.GetEnergy() > highest_energy:
                    highest_energy = track.GetEnergy()
                    mctrack = track

        return mctrack

    @staticmethod
    def get_mc_primary(frame):
        mctree = frame.Get("AntMCTree")
        primaries = mctree.GetPrimaries()
        p = None
        highest_energy = -1.

        for primary in primaries:
            if primary.GetEnergy() > highest_energy:
                highest_energy = primary.GetEnergy()
                p = primary

        return p

    @staticmethod
    def get_antares_w3(frame):
        """
        Gets weight w3 for event.
        """
        kgen_hen_year = 3.156E7
        weight_dict = frame.Get("I3MCWeightDict")
        w3 = weight_dict['GlobalWeight'] * (
            kgen_hen_year * icetray.I3Units.second
        )    # [1/yr]
        return w3

    @staticmethod
    def get_antares_w2(frame):
        """
        Gets weight w2 for event.
        """
        kgen_hen_year = 3.156E7
        weight_dict = frame.Get("I3MCWeightDict")
        # from OneWeight in units: [GeV.cm^2.s.sr]
        # to W2 in Antares units:  [GeV.m^2.(s/yr).sr]
        w2 = (
            weight_dict['OneWeight'] *
            (icetray.I3Units.centimeter2 / icetray.I3Units.meter2) *
            kgen_hen_year
        )
        return w2

    @staticmethod
    def get_nb_generated_events(frame):
        weight_dict = frame.Get("I3MCWeightDict")
        return weight_dict['NEvents']

    @staticmethod
    def get_rate(frame, omkey):

        rate = -1.

        if 'OMCondition' in frame:
            condition_map = frame['OMCondition']
            condition = condition_map[omkey]
            rate = condition.GetRate() / icetray.I3Units.kilohertz

        return rate

    def blob_generator(self):
        """
        Loop over frames in I3 file.
        Reads out geometry information of geometry frames and stores it.
        Yields physics frames.
        """
        for frame in self.fobj:

            # GEOMETRY
            if frame.Stop == icetray.I3Frame.Geometry:
                geo = frame['I3Geometry']
                self.omGeoMap = geo.omgeo
                self.cog = geo.GetDetectorCOG()

            # PHYSICS
            elif frame.Stop == icetray.I3Frame.Physics:
                yield frame

    def finish(self):
        print("finished")



# Filename: runinfo.py
"""
Prints the run table for a given detector ID.

Usage:
    runinfo [-t] DET_ID RUN
    runinfo (-h | --help)
    runinfo --version

Options:
    -t                  Show the trigger information.
    -h --help           Show this screen.
    DET_ID              Detector ID (eg. D_ARCA001).
    RUN                 Run number.

"""

import km3pipe as kp

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2018, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

log = kp.logger.get_logger(__name__)


def runinfo(run_id, det_id, show_trigger=False):
    db = kp.db.DBManager()
    df = db.run_table(det_id)
    row = df[df['RUN'] == run_id]
    if len(row) == 0:
        log.error("No database entry for run {0} found.".format(run_id))
        return
    next_row = df[df['RUN'] == (run_id + 1)]
    if len(next_row) != 0:
        end_time = next_row['DATETIME'].values[0]
        duration = (
            next_row['UNIXSTARTTIME'].values[0] -
            row['UNIXSTARTTIME'].values[0]
        ) / 1000 / 60
    else:
        end_time = duration = float('NaN')
    print("Run {0} - detector ID: {1}".format(run_id, det_id))
    print('-' * 42)
    print(
        "  Start time:         {0}\n"
        "  End time:           {1}\n"
        "  Duration [min]:     {2:.2f}\n"
        "  Start time defined: {3}\n"
        "  Runsetup ID:        {4}\n"
        "  Runsetup name:      {5}\n"
        "  T0 Calibration ID:  {6}\n".format(
            row['DATETIME'].values[0], end_time, duration,
            bool(row['STARTTIME_DEFINED'].values[0]),
            row['RUNSETUPID'].values[0], row['RUNSETUPNAME'].values[0],
            row['T0_CALIBSETID'].values[0]
        )
    )
    print("iRODS path:  {}".format(kp.tools.irods_path(det_id, run_id)))
    print("xrootd path: {}".format(kp.tools.xrootd_path(det_id, run_id)))
    if show_trigger:
        print(db.trigger_setup(row['RUNSETUPID'].values[0]))


def main():
    from docopt import docopt
    args = docopt(__doc__, version=kp.version)

    runinfo(int(args['RUN']), int(args['DET_ID']), args['-t'])



"""
Show the km3pipe etc. version used to write a H5 file.

Usage:
    h5info FILE [-r]
    h5info (-h | --help)
    h5info --version

Options:
    FILE        Input file.
    -r --raw    Dump raw metadata.
    -h --help   Show this screen.

"""

import tables

__author__ = "Moritz Lotze"
__copyright__ = "Copyright 2016, Moritz Lotze and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Moritz Lotze"
__email__ = "mlotze@km3net.de"
__status__ = "Development"


def h5info(h5name, raw=False):
    with tables.open_file(h5name) as h5:
        att_list = h5.root._v_attrs._f_list()
        if raw:
            print("HDF5 Meta Data")
            print("--------------")
            print(h5.root._v_attrs.__repr__())
            if '/header' in h5:
                print("\nHeader ('/header')")
                print("----------------")
                print(h5.get_node('/header')._v_attrs.__repr__())
            return
        if not att_list and not raw:
            print("Sorry, no metadata.")
        else:
            print("HDF5 Meta Data")
            print("--------------")
        for att in h5.root._v_attrs._f_list():
            ver = h5.root._v_attrs[att]
            print("{}: {}".format(att, ver))
        if '/header' in h5:
            print("\nHeader (/header)")
            print("----------------")
            node = h5.get_node('/header')
            for att in node._v_attrs._f_list():
                ver = node._v_attrs[att]
                print("{}: {}".format(att, ver))


def main():
    from docopt import docopt
    arguments = docopt(__doc__)

    h5info(arguments['FILE'], arguments['--raw'])



# Filename: test_foo.py
# pylint: disable=C0111,E1003,R0904,C0103,R0201,C0102
"""Make a dummy test so jenkins won't cry ):
"""
from km3pipe.testing import TestCase

__author__ = "Moritz Lotze"
__copyright__ = "Copyright 2016,  and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class TestPipeline(TestCase):
    """Tests for the main pipeline"""
    def setUp(self):
        pass

    def test_pass(self):
        assert True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:set ts=4 sts=4 sw=4 et:
"""MC Helpers.
"""

import numpy as np

from km3pipe import Module
from km3pipe.mc import pdg2name
from km3pipe.math import zenith, azimuth

__author__ = "Michael Moser and Tamas Gal and Moritz Lotze"

NEUTRINOS = {
    'nu_e',
    'anu_e',
    'nu_mu',
    'anu_mu',
    'nu_tau',
    'anu_tau',
}


class GlobalRandomState(Module):
    """Sets the global random seed of the numpy random generator

    KM3Pipe uses numpy routines exclusively to generate randomness. Setting a
    seed to a specific value will create a fully reproducible pipeline as long
    as your own modules also utilise numpy for random numbers.

    Parameters
    ----------
    seed: int, default=42

    """
    def configure(self):
        self.seed = self.get('seed', default=42)

        from numpy.random.mtrand import _rand as global_randstate
        global_randstate.seed(self.seed)

        self._random_state = np.random.RandomState(self.seed)

    @property
    def random_state(self):
        return self._random_state


class McTruth(Module):
    """Extract MC info of 1st MC track.

    Parameters
    ----------
    most_energetic_primary: bool, default=True
    """
    def configure(self):
        self.most_energetic_primary = bool(
            self.get('most_energetic_primary')
        ) or True

    @classmethod
    def t2f(cls, row):
        return pdg2name(row['type'])

    @classmethod
    def is_nu(cls, flavor):
        return flavor in NEUTRINOS

    def process(self, blob):
        mc = blob['McTracks'].conv_to('pandas')
        if self.most_energetic_primary:
            mc.sort_values('energy', ascending=False, inplace=True)
        mc = mc.head(1)
        mc['zenith'] = zenith(mc[['dir_x', 'dir_y', 'dir_z']])
        mc['azimuth'] = azimuth(mc[['dir_x', 'dir_y', 'dir_z']])
        flavor = mc.apply(self.t2f, axis=1)
        mc['is_neutrino'] = flavor.apply(self.is_nu)
        blob['McTruth'] = mc
        return blob


def convert_mc_times_to_jte_times(times_mc, evt_timestamp_in_ns, evt_mc_time):
    """
    Function that converts MC times to JTE times.

    Parameters
    ----------
    times_mc : np.ndarray
        Time array with MC times.
    evt_timestamp_in_ns : int
        Total timestamp of the event in nanoseconds.
    evt_mc_time : int
        Mc time of the event in nanoseconds.

    Returns
    -------
    ndarray
        Converted time array with JTE times.
    """
    # needs to be cast to normal ndarray (not recarray), or else we
    # would get invalid type promotion
    times_mc = np.array(times_mc).astype(float)
    times_jte = times_mc - evt_timestamp_in_ns + evt_mc_time
    return times_jte


class MCTimeCorrector(Module):
    """
    Module that converts MC hit times to JTE times.
    Thus, the following tables need to be converted:
    - mc_tracks
    - mc_hits

    Parameters
    ----------
    mc_hits_key : str, optional
        Name of the mc_hits to convert (default: 'McHits').
    mc_tracks_key : str, optional
        Name of the mc_tracks to convert (default: 'McTracks').
    event_info_key : str, optional
        Name of the event_info to store this in (default: 'EventInfo').
    """
    def configure(self):
        # get event_info, hits and mc_tracks key ; define conversion func
        self.event_info_key = self.get('event_info_key', default='EventInfo')
        self.mc_tracks_key = self.get('mc_tracks_key', default='McTracks')
        self.mc_hits_key = self.get('mc_hits_key', default='McHits')

        self.convert_mc_times_to_jte_times = \
            np.frompyfunc(convert_mc_times_to_jte_times, 3, 1)

    def process(self, blob):
        # convert the mc times to jte times
        event_info = blob[self.event_info_key]
        mc_tracks = blob[self.mc_tracks_key]
        mc_hits = blob[self.mc_hits_key]
        timestamp_in_ns = event_info.timestamp * 1e9 + event_info.nanoseconds

        mc_tracks['time'] = self.convert_mc_times_to_jte_times(
            mc_tracks.time, timestamp_in_ns, event_info.mc_time
        )
        mc_hits['time'] = self.convert_mc_times_to_jte_times(
            mc_hits.time, timestamp_in_ns, event_info.mc_time
        )

        blob[self.mc_tracks_key] = mc_tracks
        blob[self.mc_hits_key] = mc_hits

        return blob


def slew(tot):
    """Calculate the time slewing of a PMT response for a given ToT"""

    #p0 = 7.70824
    #p1 = 0.00879447
    #p2 = -0.0621101
    #p3 = -1.90226

    p0 = 13.6488662517
    p1 = -0.128744123166
    p2 = -0.0174837749244
    p3 = -4.47119633965

    corr = p0 * np.exp(p1 * np.sqrt(tot) + p2 * tot) + p3
    return corr



# Filename: test_clb.py
# pylint: disable=C0111,R0904,C0103
"""
...

"""
from io import StringIO
from os.path import join, dirname

from km3pipe.testing import TestCase, data_path
from km3pipe.io.clb import CLBPump


__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class TestCLBPump(TestCase):
    def setUp(self):
        self.pump = CLBPump(file=data_path('daq/clb.dqd'))

    def test_determine_packet_positions_finds_packets(self):
        self.assertListEqual([0, 1406, 2284], self.pump._packet_positions)

    def test_length(self):
        assert 3 == len(self.pump)

    def test_getindex(self):
        blob = self.pump[0]
        self.assertEqual(0, blob['PacketInfo'].run[0])
        self.assertEqual('TTDC', blob['PacketInfo'].data_type[0])
        assert 229 == len(blob['Hits'])
        a_pmt_data = blob['Hits'][2]
        self.assertEqual(15, a_pmt_data.channel_id)
        self.assertEqual(66541067, a_pmt_data.time)
        self.assertEqual(28, a_pmt_data.tot)

        assert 229 == len(self.pump[0]['Hits'])
        assert 141 == len(self.pump[1]['Hits'])
        assert 229 == len(self.pump[2]['Hits'])

    def test_iterator(self):
        i = 0
        for b in self.pump:
            i += 1
        assert 3 == i

        i = 0
        for b in self.pump:
            i += 1
        assert 3 == i



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ts=4 sw=4 et
"""
===========
Trigger Map
===========

This script creates a histogram to show the trigger contribution for events.

.. code-block:: bash

    Usage:
        trigger_map.py [-d DU -p PLOT_FILENAME] FILENAME
        trigger_map.py --version

    Option:
        FILENAME          Name of the input file.
        -d DU             Only plot for the given DU.
        -p PLOT_FILENAME  The filename of the plot [default: trigger_map.png].
        -h --help         Show this screen.

"""

from docopt import docopt
import matplotlib
# Force matplotlib to not use any Xwindows backend.
matplotlib.use('Agg')    # noqa
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
import numpy as np

import km3pipe as kp
import km3pipe.style


class TriggerMap(kp.Module):
    """Creates a plot to show the number of triggered hits for each DOM."""
    def configure(self):
        self.det = self.require('detector')
        self.plot_filename = self.require('plot_filename')
        self.subtitle = self.get('subtitle', default='')
        self.du = self.get('du')
        if self.du is not None:
            self.n_dus = 1
            self.n_doms = 18
        else:
            self.n_dus = self.det.n_dus
            self.n_doms = int(self.det.n_doms / self.n_dus)
        self.hit_counts = []

    def process(self, blob):
        hits = blob['Hits'].triggered_rows
        dom_ids = np.unique(hits.dom_id)
        hit_counts = np.zeros(self.n_dus * self.n_doms)
        for dom_id in dom_ids:
            n_hits = np.sum(hits.dom_id == dom_id)
            du, floor, _ = self.det.doms[dom_id]
            if self.du is not None and du != self.du:
                continue
            if self.du:
                du = 1
            hit_counts[(du - 1) * self.n_doms + floor - 1] += n_hits
        self.hit_counts.append(hit_counts)

        return blob

    def finish(self):
        self.create_plot()

    def create_plot(self):
        km3pipe.style.use("km3pipe")

        title = "Trigger Map\n{}".format(self.subtitle)
        fig, ax = plt.subplots(figsize=(16, 8))
        ax.grid(True)
        ax.set_axisbelow(True)
        hit_mat = np.array([np.array(x) for x in self.hit_counts]).transpose()
        im = ax.matshow(
            hit_mat,
            interpolation='nearest',
            filternorm=None,
            cmap='plasma',
            aspect='auto',
            origin='lower',
            zorder=3,
            norm=LogNorm(vmin=1, vmax=np.amax(hit_mat))
        )
        yticks = np.arange(self.n_doms * self.n_dus)
        ytick_label_templ = "DU{0:.0f}-DOM{1:02d}" if self.du else "DOM{1:02d}"
        ytick_labels = [
            ytick_label_templ.format(
                np.ceil((y + 1) / self.n_doms), y % (self.n_doms) + 1
            ) for y in yticks
        ]
        ax.set_yticks(yticks)
        ax.set_yticklabels(ytick_labels)
        ax.tick_params(labelbottom=True)
        ax.tick_params(labeltop=True)
        ax.set_xlabel("event id")
        ax.set_title(title)
        cb = fig.colorbar(im, pad=0.05)
        cb.set_label("number of triggered hits")

        fig.tight_layout()

        plt.savefig(self.plot_filename, dpi=120, bbox_inches="tight")


if __name__ == '__main__':
    args = docopt(__doc__, version='1.0')
    print(args)
    du = int(args['-d']) if args['-d'] else None
    det = kp.hardware.Detector(det_id=29)
    pipe = kp.Pipeline()
    pipe.attach(kp.io.jpp.EventPump, filename=args['FILENAME'])
    pipe.attach(
        TriggerMap,
        detector=det,
        du=du,
        plot_filename=args['-p'],
        subtitle=args['FILENAME']
    )
    pipe.drain()
